@@ -8,7 +8,7 @@
 
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2013 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2017 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -31,7 +31,7 @@
  *   You should have received a copy of the GNU General Public License along
  *   with FORM.  If not, see <http://www.gnu.org/licenses/>.
  */
-/* #] License : */ 
+/* #] License : */
  
 /**
  *	The next macros were introduced when TFORM was programmed. In the case of
@@ -47,17 +47,6 @@
  *	These macro's contain the comma as well. Hence we need special macro's if
  *	there are no other arguments. These are called PHEAD0 and BHEAD0.
  */
-#ifdef WITHPTHREADS
-#define PHEAD  ALLPRIVATES *B,
-#define PHEAD0 ALLPRIVATES *B
-#define BHEAD  B,
-#define BHEAD0 B
-#else
-#define PHEAD
-#define PHEAD0 VOID
-#define BHEAD
-#define BHEAD0
-#endif
  
 #define WITHOUTERROR 0
 #define WITHERROR 1
@@ -75,6 +64,7 @@
 #define PREREADSTREAM2 6
 #define EXTERNALCHANNELSTREAM 7
 #define PREREADSTREAM3 8
+#define REVERSEFILESTREAM 9
  
 #define ENDOFSTREAM 0xFF
 #define ENDOFINPUT 0xFF
@@ -105,6 +95,7 @@
 #define NOPARALLEL_DOLLAR       0x0001
 #define NOPARALLEL_RHS          0x0002
 #define NOPARALLEL_CONVPOLY     0x0004
+#define NOPARALLEL_SPECTATOR    0x0008
 #define NOPARALLEL_USER         0x0010
 #define NOPARALLEL_TBLDOLLAR    0x0100
 #define NOPARALLEL_NPROC        0x0200
@@ -309,6 +300,9 @@ typedef int (*TFUN1)();
 #define DOUBLEFORTRANMODE 33
 #define DOUBLEPRECISIONFLAG 32
 #define NODOUBLEMASK 31
+#define QUADRUPLEFORTRANMODE 65
+#define QUADRUPLEPRECISIONFLAG 64
+#define NOQUADMASK 63
 #define NORMALFORMAT 0
 #define NOSPACEFORMAT 1
 
@@ -353,6 +347,7 @@ typedef int (*TFUN1)();
 #define HAAKJE0 9
 #define FUNCTION 20
 
+#define TMPPOLYFUN 14
 #define ARGFIELD 15
 #define SNUMBER 16
 #define LNUMBER 17
@@ -447,8 +442,19 @@ typedef int (*TFUN1)();
 #define EXTEUCLIDEAN 98
 #define MAKERATIONAL 99
 #define INVERSEFUNCTION 100
+#define IDFUNCTION 101
+#define PUTFIRST 102
+#define PERMUTATIONS 103
+#define PARTITIONS 104
+#define MULFUNCTION 105
+#define TOPOLOGIES 106
+#define DIAGRAMS 107
+#define VERTEX 108
+#define EDGE 109
+/*#define ALLWILD 109 ???? */
+#define SIZEOFFUNCTION 110
 
-#define MAXBUILTINFUNCTION 100
+#define MAXBUILTINFUNCTION 110
 #define FIRSTUSERFUNCTION 150
 
 /*
@@ -553,6 +559,12 @@ typedef int (*TFUN1)();
 #define TYPEENDDOLOOP 75
 #define TYPEDROPSYMBOLS 76
 #define TYPEPUTINSIDE 77
+#define TYPETOSPECTATOR 78
+#define TYPEARGTOEXTRASYMBOL 79
+#define TYPECANONICALIZE 80
+#define TYPESWITCH 81
+#define TYPEENDSWITCH 82
+
 /*
 	The codes for the 'operations' that are part of TYPEOPERATION.
 */
@@ -603,8 +615,10 @@ typedef int (*TFUN1)();
 #define CLEANFLAG 0
 #define DIRTYFLAG 1
 #define DIRTYSYMFLAG 2
+#define MUSTCLEANPRF 4
+#define SUBTERMUSED1 8
+#define SUBTERMUSED2 16
 #define ALLDIRTY (DIRTYFLAG|DIRTYSYMFLAG)
-#define CLEANPRF 4
 
 #define ARGHEAD 2
 #define FUNHEAD 3
@@ -646,6 +660,10 @@ typedef int (*TFUN1)();
 #define UNHIDEGEXPRESSION 16
 #define INTOHIDELEXPRESSION 17
 #define INTOHIDEGEXPRESSION 18
+#define SPECTATOREXPRESSION 19
+#define DROPSPECTATOREXPRESSION 20
+#define SKIPUNHIDELEXPRESSION 21
+#define SKIPUNHIDEGEXPRESSION 22
 
 #define PRINTOFF 0
 #define PRINTON 1
@@ -693,6 +711,7 @@ typedef int (*TFUN1)();
 #define INDEX_       9   /* index only */
 #define Q_          10   /* rational */
 #define DUMMYINDEX_ 11   /* dummy index only */
+#define VECTOR_     12   /* vector only */
 
 /*
 	Special indices.
@@ -756,14 +775,18 @@ typedef int (*TFUN1)();
 #define SUBONCE 2
 #define SUBONLY 3
 #define SUBMANY 4
-#define SUBALL 5
+#define SUBVECTOR 5
 #define SUBSELECT 6
+#define SUBALL 7
 #define SUBMASK 15
 #define SUBDISORDER 16
 #define SUBAFTER 32
 #define SUBAFTERNOT 64
 
-#define IDHEAD 5
+#define IDHEAD 6
+
+#define DOLLARFLAG 1
+#define NORMALIZEFLAG 2
 
 #define GIDENT 1
 #define GFIVE 4
@@ -783,6 +806,7 @@ typedef int (*TFUN1)();
 #define IFEXPRESSION 7
 #define IFDOLLAREXTRA 8
 #define IFISFACTORIZED 9
+#define IFOCCURS 10
 #define GREATER	0
 #define GREATEREQUAL 1
 #define LESS 2
@@ -945,6 +969,11 @@ typedef int (*TFUN1)();
 #define ISLYNDONR   14
 #define TOLYNDON    15
 #define TOLYNDONR   16
+#define ADDARG      17
+#define MULTIPLYARG 18
+#define DROPARG     19
+#define SELECTARG   20
+#define DEDUPARG    21
 
 #define BASECODE 1
 #define YESLYNDON 1
@@ -970,6 +999,7 @@ typedef int (*TFUN1)();
 
 #define O_OCCURRENCE 0
 #define O_MCTS 1
+#define O_SIMULATED_ANNEALING 2
 
 #define O_FORWARD 0
 #define O_BACKWARD 1
@@ -979,3 +1009,39 @@ typedef int (*TFUN1)();
 #define OPTHEAD 3
 #define DOALL 1
 #define ONLYFUNCTIONS 2
+
+#define INUSE 1
+#define COULDCOMMUTE 2
+#define DOESNOTCOMMUTE 4
+ 
+#define DICT_NONUMBERS 0
+#define DICT_INTEGERONLY 1
+#define DICT_RATIONALONLY 2
+#define DICT_ALLNUMBERS 3
+#define DICT_NOVARIABLES 0
+#define DICT_DOVARIABLES 1
+#define DICT_NOSPECIALS 0
+#define DICT_DOSPECIALS 1
+#define DICT_NOFUNWITHARGS 0
+#define DICT_DOFUNWITHARGS 1
+#define DICT_NOTINDOLLARS 0
+#define DICT_INDOLLARS 1
+
+#define DICT_INTEGERNUMBER 1
+#define DICT_RATIONALNUMBER 2
+#define DICT_SYMBOL 3
+#define DICT_VECTOR 4
+#define DICT_INDEX 5
+#define DICT_FUNCTION 6
+#define DICT_FUNCTION_WITH_ARGUMENTS 7
+#define DICT_SPECIALCHARACTER 8
+#define DICT_RANGE 9
+
+#define READSPECTATORFLAG 3
+#define GLOBALSPECTATORFLAG 1
+
+#define ORDEREDSET 1
+
+#define DENSETABLE 1
+#define SPARSETABLE 0
+
