@@ -4,7 +4,7 @@
  */
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2013 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2017 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -167,6 +167,106 @@ getnumber:;
 
 /*
  		#] GetIfDollarNum : 
+ 		#[ FindVar :
+*/
+
+int FindVar(WORD *v, WORD *term)
+{
+	WORD *t, *tstop, *m, *mstop, *f, *fstop, *a, *astop;
+	GETSTOP(term,tstop);
+	t = term+1;
+	while ( t < tstop ) {
+		if ( *v == *t && *v < FUNCTION ) {	/* VECTOR, INDEX, SYMBOL, DOTPRODUCT */
+			switch ( *v ) {
+				case SYMBOL:
+					m = t+2; mstop = t+t[1];
+					while ( m < mstop ) {
+						if ( *m == v[1] ) return(1);
+						m += 2;
+					}
+					break;
+				case INDEX:
+				case VECTOR:
+InVe:
+					m = t+2; mstop = t+t[1];
+					while ( m < mstop ) {
+						if ( *m == v[1] ) return(1);
+						m++;
+					}
+					break;
+				case DOTPRODUCT:
+					m = t+2; mstop = t+t[1];
+					while ( m < mstop ) {
+						if ( *m == v[1] && m[1] == v[2] ) return(1);
+						if ( *m == v[2] && m[1] == v[1] ) return(1);
+						m += 3;
+					}
+					break;
+			}
+		}
+		else if ( *v == VECTOR && *t == INDEX ) goto InVe;
+		else if ( *v == INDEX && *t == VECTOR ) goto InVe;
+		else if ( ( *v == VECTOR || *v == INDEX ) && *t == DOTPRODUCT ) {
+			m = t+2; mstop = t+t[1];
+			while ( m < mstop ) {
+				if ( v[1] == m[0] || v[1] == m[1] ) return(1);
+				m += 3;
+			}
+		}
+		else if ( *t >= FUNCTION ) {
+			if ( *v == FUNCTION && v[1] == *t ) return(1);
+			if ( functions[*t-FUNCTION].spec > 0 ) {
+			  if ( *v == VECTOR || *v == INDEX ) { /* we need to check arguments */
+				int i;
+				for ( i = FUNHEAD; i < t[1]; i++ ) {
+					if ( v[1] == t[i] ) return(1);
+				}
+			  }
+			}
+			else {
+			  fstop = t + t[1]; f = t + FUNHEAD;
+			  while ( f < fstop ) { /* Do the arguments one by one */
+				if ( *f <= 0 ) {
+					switch ( *f ) {
+						case -SYMBOL:
+							if ( *v == SYMBOL && v[1] == f[1] ) return(1);
+							f += 2;
+							break;
+						case -VECTOR:
+						case -MINVECTOR:
+						case -INDEX:
+							if ( ( *v == VECTOR || *v == INDEX )
+							 && ( v[1] == f[1] ) ) return(1);
+							f += 2;
+							break;
+						case -SNUMBER:
+							f += 2;
+							break;
+						default:
+							if ( *v == FUNCTION && v[1] == -*f && *f <= -FUNCTION ) return(1);
+							if ( *f <= -FUNCTION ) f++;
+							else f += 2;
+							break;
+					}
+				}
+				else {
+					a = f + ARGHEAD; astop = f + *f;
+					while ( a < astop ) {
+						if ( FindVar(v,a) == 1 ) return(1);
+						a += *a;
+					}
+					f = astop;
+				}
+			  }
+			}
+		}
+		t += t[1];
+	}
+	return(0);
+}
+
+/*
+ 		#] FindVar : 
  		#[ DoIfStatement :				WORD DoIfStatement(PHEAD ifcode,term)
 
 		The execution time part of the if-statement.
@@ -254,6 +354,17 @@ WORD DoIfStatement(PHEAD WORD *ifcode, WORD *term)
 					r++;
 				}
 				return(1);
+			case IFOCCURS:
+				{
+					WORD *OccStop = ifp + ifp[1];
+					ifp += 2;
+					while ( ifp < OccStop ) {
+						if ( FindVar(ifp,term) == 1 ) return(1);
+						if ( *ifp == DOTPRODUCT ) ifp += 3;
+						else ifp += 2;
+					}
+				}
+				return(0);
 			default:
 /*
 				Now we have a subexpression. Test first for one with a single item.
@@ -436,6 +547,7 @@ WORD DoIfStatement(PHEAD WORD *ifcode, WORD *term)
 								break;
 							}
 							d->index = d->where[2];
+							/* fall through */
 						case DOLINDEX:
 							if ( d->index == 0 ) {
 								ncoef2 = coef2[0] = 0; coef2[1] = 1;
@@ -471,6 +583,7 @@ WORD DoIfStatement(PHEAD WORD *ifcode, WORD *term)
 								ncoef2 = coef2[0] = 0; coef2[1] = 1;
 								break;
 							}
+							/* fall through */
 						case DOLARGUMENT:
 							if ( d->where[0] == -SNUMBER ) {
 								if ( d->where[1] == 0 ) {
@@ -581,6 +694,21 @@ generic:;
 				coef2[0] = ncoef2;
 				coef2[1] = 1;
 				break;
+			case IFOCCURS:
+				{
+					WORD *OccStop = ifp + ifp[1], *ifpp = ifp+2;
+					ncoef2 = 0;
+					while ( ifpp < OccStop ) {
+						if ( FindVar(ifpp,term) == 1 ) {
+							ncoef2 = 1; break;
+						}
+						if ( *ifpp == DOTPRODUCT ) ifp += 3;
+						else ifpp += 2;
+					}
+					coef2[0] = ncoef2;
+					coef2[1] = 1;
+				}
+				break;
 			default:
 				break;
 		}
@@ -772,7 +900,7 @@ WORD HowMany(PHEAD WORD *ifcode, WORD *term)
 				FindOnly(BHEAD newterm,m) ) ) {
 				Substitute(BHEAD newterm,m,1);
 				if ( numdollars ) {
-					WildDollars(BHEAD0);
+					WildDollars(BHEAD (WORD *)0);
 					numdollars = 0;
 				}
 				ClearWild(BHEAD0);
@@ -796,7 +924,7 @@ WORD HowMany(PHEAD WORD *ifcode, WORD *term)
 					do {
 						Substitute(BHEAD newterm,m,1);
 						if ( numdollars ) {
-							WildDollars(BHEAD0);
+							WildDollars(BHEAD (WORD *)0);
 							numdollars = 0;
 						}
 						ClearWild(BHEAD0);
@@ -808,7 +936,7 @@ WORD HowMany(PHEAD WORD *ifcode, WORD *term)
 					do {
 						Substitute(BHEAD newterm,m,1);
 						if ( numdollars ) {
-							WildDollars(BHEAD0);
+							WildDollars(BHEAD (WORD *)0);
 							numdollars = 0;
 						}
 						ClearWild(BHEAD0);
@@ -821,7 +949,7 @@ WORD HowMany(PHEAD WORD *ifcode, WORD *term)
 					do {
 						Substitute(BHEAD newterm,m,1);
 						if ( numdollars ) {
-							WildDollars(BHEAD0);
+							WildDollars(BHEAD (WORD *)0);
 							numdollars = 0;
 						}
 						ClearWild(BHEAD0);
@@ -843,7 +971,7 @@ WORD HowMany(PHEAD WORD *ifcode, WORD *term)
 			if ( FindRest(BHEAD newterm,m) && ( AN.UsedOtherFind || FindOnce(BHEAD newterm,m) ) ) {
 				Substitute(BHEAD newterm,m,1);
 				if ( numdollars ) {
-					WildDollars(BHEAD0);
+					WildDollars(BHEAD (WORD *)0);
 					numdollars = 0;
 				}
 				ClearWild(BHEAD0);
@@ -854,7 +982,7 @@ WORD HowMany(PHEAD WORD *ifcode, WORD *term)
 		case SUBMULTI :
 			RetVal = FindMulti(BHEAD term,m);
 			break;
-		case SUBALL :
+		case SUBVECTOR :
 			RetVal = 0;
 			for ( i = 0; i < *term; i++ ) ww[i] = term[i];			
 			while ( ( power = FindAll(BHEAD ww,m,AR.Cnumlhs,ifcode) ) != 0 ) { RetVal += power; }
@@ -913,5 +1041,161 @@ VOID DoubleIfBuffers()
 /*
  		#] DoubleIfBuffers : 
   	#] If statement : 
+  	#[ Switch statement :
+ 		#[ DoSwitch :
 */
 
+int DoSwitch(PHEAD WORD *term, WORD *lhs)
+{
+/*
+	For the moment we ignore the compiler buffer problems.
+*/
+	WORD numdollar = lhs[2];
+	WORD ncase = DolToNumber(BHEAD numdollar);
+	SWITCHTABLE *swtab = FindCase(lhs[3],ncase);
+	return(Generator(BHEAD term,swtab->value));
+}
+
+/*
+ 		#] DoSwitch : 
+ 		#[ DoEndSwitch :
+*/
+
+int DoEndSwitch(PHEAD WORD *term, WORD *lhs)
+{
+	SWITCH *sw = AC.SwitchArray+lhs[2];
+	return(Generator(BHEAD term,sw->endswitch.value+1));
+}
+
+/*
+ 		#] DoEndSwitch : 
+ 		#[ FindCase :
+*/
+
+SWITCHTABLE *FindCase(WORD nswitch, WORD ncase)
+{
+/*
+	First find the switch table and determine how we have to search.
+*/
+	SWITCH *sw = AC.SwitchArray+nswitch;
+	WORD hi, lo, med;
+	if ( sw->typetable == DENSETABLE ) {
+		med = ncase - sw->caseoffset;
+		if ( med >= sw->numcases || med < 0 ) return(&sw->defaultcase);
+	}
+	else {
+/*
+		We need a binary search in the table.
+*/
+		if ( ncase > sw->maxcase || ncase < sw->mincase ) return(&sw->defaultcase);
+		hi = sw->numcases-1; lo = 0;
+		for(;;) {
+			med = (hi+lo)/2;
+			if ( ncase == sw->table[med].ncase ) break;
+            else if ( ncase > sw->table[med].ncase ) {
+				lo = med+1;
+				if ( lo > hi ) return(&sw->defaultcase);
+			}
+			else {
+				hi = med-1;
+				if ( hi < lo ) return(&sw->defaultcase);
+			}
+		}
+	}
+	return(&sw->table[med]);
+}
+
+/*
+ 		#] FindCase : 
+ 		#[ DoubleSwitchBuffers :
+*/
+
+int DoubleSwitchBuffers()
+{
+	int newmax, i;
+	SWITCH *newarray;
+	WORD *newheap;
+	if ( AC.MaxSwitch == 0 ) newmax = 10;
+	else                     newmax = 2*AC.MaxSwitch;
+	newarray = (SWITCH *)Malloc1(sizeof(SWITCH)*(newmax+1),"SwitchArray");
+	newheap = (WORD *)Malloc1(sizeof(WORD)*(newmax+1),"SwitchHeap");
+	if ( AC.MaxSwitch ) {
+		for ( i = 0; i < AC.MaxSwitch; i++ ) {
+			newarray[i] = AC.SwitchArray[i];
+			newheap[i] = AC.SwitchHeap[i];
+		}
+		M_free(AC.SwitchHeap,"AC.SwitchHeap");
+		M_free(AC.SwitchArray,"AC.SwitchArray");
+	}
+	for ( i = AC.MaxSwitch; i <= newmax; i++ ) {
+		newarray[i].table = 0;
+		newarray[i].tablesize = 0;
+		newarray[i].defaultcase.ncase = 0;
+		newarray[i].defaultcase.value = 0;
+		newarray[i].defaultcase.compbuffer = 0;
+		newarray[i].endswitch.ncase = 0;
+		newarray[i].endswitch.value = 0;
+		newarray[i].endswitch.compbuffer = 0;
+		newarray[i].typetable = 0;
+		newarray[i].mincase = 0;
+		newarray[i].maxcase = 0;
+		newarray[i].numcases = 0;
+		newarray[i].caseoffset = 0;
+		newarray[i].iflevel = 0;
+		newarray[i].whilelevel = 0;
+		newarray[i].nestingsum = 0;
+		newheap[i] = 0;
+	}
+	AC.SwitchArray = newarray;
+	AC.SwitchHeap = newheap;
+	AC.MaxSwitch = newmax;
+	return(0);
+}
+
+/*
+ 		#] DoubleSwitchBuffers : 
+ 		#[ SwitchSplitMerge :
+
+		Sorts an array of WORDs. No adding of equal objects.
+*/
+
+VOID SwitchSplitMergeRec(SWITCHTABLE *array,WORD num,SWITCHTABLE *auxarray)
+{
+	WORD n1,n2,i,j,k;
+	SWITCHTABLE *t1,*t2, t;
+	if ( num < 2 ) return;
+	if ( num == 2 ) {
+		if ( array[0].ncase > array[1].ncase ) {
+			t = array[0]; array[0] = array[1]; array[1] = t;
+		}
+		return;
+	}
+	n1 = num/2;
+	n2 = num - n1;
+	SwitchSplitMergeRec(array,n1,auxarray);
+	SwitchSplitMergeRec(array+n1,n2,auxarray);
+	if ( array[n1-1].ncase <= array[n1].ncase ) return;
+
+	t1 = array; t2 = auxarray; i = n1; NCOPY(t2,t1,i);
+	i = 0; j = n1; k = 0;
+	while ( i < n1 && j < num ) {
+		if ( auxarray[i].ncase <= array[j].ncase ) { array[k++] = auxarray[i++]; }
+		else { array[k++] = array[j++]; }
+	}
+	while ( i < n1 ) array[k++] = auxarray[i++];
+/*
+	Remember: remnants of j are still in place!
+*/
+}
+
+VOID SwitchSplitMerge(SWITCHTABLE *array,WORD num)
+{
+	SWITCHTABLE *auxarray = (SWITCHTABLE *)Malloc1(sizeof(SWITCHTABLE)*num/2,"SwitchSplitMerge");
+	SwitchSplitMergeRec(array,num,auxarray);
+	M_free(auxarray,"SwitchSplitMerge");
+}
+
+/*
+ 		#] SwitchSplitMerge : 
+  	#] Switch statement : 
+*/
