@@ -8,7 +8,7 @@
  */
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2013 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2017 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -328,6 +328,102 @@ NotFound:;
 
 /*
   	#] GetName : 
+  	#[ GetFunction :
+
+	Gets either a function or a $ that should expand into a function
+	during runtime. In the case of the $ the value in funnum is -dolnum-1.
+	The return value is the position after the name of the function or the $.
+*/
+ 
+static WORD one = 1;
+
+UBYTE *GetFunction(UBYTE *s,WORD *funnum)
+{
+	int type;
+	WORD numfun;
+	UBYTE *t1, c;
+	if ( *s == '$' ) {
+		t1 = s+1; while ( FG.cTable[*t1] < 2 ) t1++;
+		c = *t1; *t1 = 0;
+		if ( ( type = GetName(AC.dollarnames,s+1,&numfun,NOAUTO) ) == CDOLLAR ) {
+			*funnum = -numfun-2;
+		}
+		else {
+			MesPrint("&%s is undefined",s);
+			numfun = AddDollar(s+1,DOLINDEX,&one,1);
+			*funnum = 0;
+		}
+	}
+	else {
+		t1 = SkipAName(s);
+		c = *t1; *t1 = 0;
+		if ( ( ( type = GetName(AC.varnames,s,&numfun,WITHAUTO) ) != CFUNCTION )
+			|| ( functions[numfun].spec != 0 ) ) {
+			MesPrint("&%s should be a regular function",s);
+			*funnum = 0;
+			if ( type < 0 ) {
+				if ( GetName(AC.exprnames,s,&numfun,NOAUTO) == NAMENOTFOUND )
+				AddFunction(s,0,0,0,0,0,-1,-1);
+			}
+			*t1 = c;
+			return(t1);
+		}
+		*funnum = numfun+FUNCTION;
+	}
+	*t1 = c;
+	return(t1);
+}
+
+/*
+  	#] GetFunction : 
+  	#[ GetNumber :
+
+	Gets either a number or a $ that should expand into a number
+	during runtime. In the case of the $ the value in num is -dolnum-2.
+	The return value is the position after the number or the $.
+*/
+
+UBYTE *GetNumber(UBYTE *s,WORD *num)
+{
+	int type;
+	WORD numfun;
+	UBYTE *t1, c;
+	while ( *s == '+' ) s++;
+	if ( *s == '$' ) {
+		t1 = s+1; while ( FG.cTable[*t1] < 2 ) t1++;
+		c = *t1; *t1 = 0;
+		if ( ( type = GetName(AC.dollarnames,s+1,&numfun,NOAUTO) ) == CDOLLAR ) {
+			*num = -numfun-2;
+		}
+		else {
+			MesPrint("&%s is undefined",s);
+			numfun = AddDollar(s+1,DOLINDEX,&one,1);
+			*num = -1;
+		}
+	}
+	else if ( *s >= '0' && *s <= '9' ) {
+		ULONG x = *s++ - '0';
+		while ( *s >= '0' && *s <= '9' ) { x = 10*x + (*s++-'0'); }
+		t1 = s;
+		if ( x >= MAXPOSITIVE ) goto illegal;
+		*num = (WORD)x;
+		return(t1);
+	}
+	else {
+		if ( *s == '-' ) { s++; }
+		if ( *s >= '0' && *s <= '9' ) { while ( *s >= '0' && *s <= '9' ) s++; t1 = s; }
+		else { t1 = SkipAName(s); }
+illegal:
+		*num = -1;
+		MesPrint("&Illegal option in Canonicalize statement. Should be a nonnegative number or $ variable.");
+		return(t1);
+	}
+	*t1 = c;
+	return(t1);
+}
+
+/*
+  	#] GetNumber : 
   	#[ GetLastExprName :
 
 	When AutoDeclare is an active statement.
@@ -829,6 +925,7 @@ int AddSymbol(UBYTE *name, int minpow, int maxpow, int cplx, int dim)
 	sym->minpower = minpow;
 	sym->maxpower = maxpow;
 	sym->complex  = cplx;
+	sym->flags    = 0;
 	sym->node     = nodenum;
 	sym->dimension= dim;
 	while ( *s ) s++;
@@ -868,12 +965,13 @@ IllForm:	MesPrint("&Illegally formed name in symbol statement");
 			     if ( tolower(*s) == 'r' ) cplx = VARTYPENONE;
 			else if ( tolower(*s) == 'c' ) cplx = VARTYPECOMPLEX;
 			else if ( tolower(*s) == 'i' ) cplx = VARTYPEIMAGINARY;
-			else if ( ( ( *s == '-' || *s == '+' ) && ( s[1] >= '0' && s[1] <= '9' ) )
+			else if ( ( ( *s == '-' || *s == '+' || *s == '=' )
+				 && ( s[1] >= '0' && s[1] <= '9' ) )
 				 || ( *s >= '0' && *s <= '9' ) ) {
 				LONG x;
 				sgn = 0;
 				if ( *s == '-' ) { sgn = VARTYPEMINUS; s++; }
-				else if ( *s == '+' ) { sgn = 0; s++; }
+				else if ( *s == '+' || *s == '=' ) { sgn = 0; s++; }
 				x = *s -'0';
 				while ( s[1] >= '0' && s[1] <= '9' ) {
 					x = 10*x + (s[1] - '0'); s++;
@@ -958,6 +1056,11 @@ skippar:		error = 1;
 			if ( type != CSYMBOL ) error = NameConflict(type,name);
 			else {
 				SYMBOLS sym = (SYMBOLS)(AC.Symbols->lijst) + numsymbol;
+				if ( ( numsymbol == AC.lPolyFunVar ) && ( AC.lPolyFunType > 0 )
+					 && ( AC.lPolyFun != 0 ) && ( minpow > -MAXPOWER || maxpow < MAXPOWER ) ) {
+					MesPrint("&The symbol %s is used by power expansions in the PolyRatFun!",name);
+					error = 1;
+				}
 				sym->complex  = cplx;
 				sym->minpower = minpow;
 				sym->maxpower = maxpow;
@@ -989,6 +1092,7 @@ int AddIndex(UBYTE *name, int dim, int dim4)
 	ind->name      = AddName(*AC.activenames,name,CINDEX,numindex,&nodenum);
 	ind->type      = 0;
 	ind->dimension = dim;
+	ind->flags     = 0;
 	ind->nmin4     = dim4;
 	ind->node      = nodenum;
 	while ( *s ) s++;
@@ -1142,6 +1246,7 @@ int AddVector(UBYTE *name, int cplx, int dim)
 	v->complex = cplx;
 	v->node    = nodenum;
 	v->dimension = dim;
+	v->flags = 0;
 	while ( *s ) s++;
 	v->namesize = (s-name)+1;
 	return(numvector);
@@ -1224,6 +1329,7 @@ int AddFunction(UBYTE *name, int comm, int istensor, int cplx, int symprop, int
 	fun->spec = istensor;
 	fun->complex  = cplx;
 	fun->tabl = 0;
+	fun->flags = 0;
 	fun->node = nodenum;
 	fun->symminfo = 0;
 	fun->symmetric = symprop;
@@ -1237,6 +1343,94 @@ int AddFunction(UBYTE *name, int comm, int istensor, int cplx, int symprop, int
 
 /*
   	#] AddFunction : 
+  	#[ CoCommuteInSet :
+
+	Commuting,f1,...,fn;
+*/
+
+int CoCommuteInSet(UBYTE *s)
+{
+	UBYTE *name, *ss, c, *start = s;
+	WORD number, type, *g, *gg;
+	int error = 0, i, len = StrLen(s), len2 = 0;
+	if ( AC.CommuteInSet != 0 ) {
+		g = AC.CommuteInSet;
+		while ( *g ) g += *g;
+		len2 = g - AC.CommuteInSet;
+		if ( len2+len+3 > AC.SizeCommuteInSet ) {
+			gg = (WORD *)Malloc1((len2+len+3)*sizeof(WORD),"CommuteInSet");
+			for ( i = 0; i < len2; i++ ) gg[i] = AC.CommuteInSet[i];
+			gg[len2] = 0;
+			M_free(AC.CommuteInSet,"CommuteInSet");
+			AC.CommuteInSet = gg;
+			AC.SizeCommuteInSet = len+len2+3;
+			g = AC.CommuteInSet+len2;
+		}
+	}
+	else {
+		AC.SizeCommuteInSet = len+2;
+		g = AC.CommuteInSet = (WORD *)Malloc1((len+3)*sizeof(WORD),"CommuteInSet");
+		*g = 0;
+	}
+	gg = g++;
+	ss = s-1;
+	for(;;) {
+		while ( *s == ',' || *s == ' ' || *s == '\t' ) s++;
+		if ( *s == 0 ) {
+			if ( s - start >= len ) break;
+			*s = '}'; s++;
+			*g = 0;
+			*gg = g-gg;
+			if ( *gg < 2 ) {
+				MesPrint("&There should be at least two noncommuting functions or tensors in a commuting statement.");
+				error = 1;		
+			}
+			else if ( *gg == 2 ) {
+				gg[2] = gg[1]; gg[3] = 0; gg[0] = 3;
+			}
+			gg = g++;
+			continue;
+		}
+		if ( s > ss ) {
+			if ( *s != '{' ) {
+				MesPrint("&The CommuteInSet statement should have sets enclosed in {}.");
+				error = 1;
+				break;
+			}
+			ss = s;
+			SKIPBRA2(ss) /* Note that parentheses were tested before */
+			*ss = 0;
+			s++;
+		}
+		name = s;
+		s = SkipAName(s);
+		c = *s; *s = 0;
+		if ( ( type = GetName(AC.varnames,name,&number,NOAUTO) ) != CFUNCTION ) {
+			MesPrint("&%s is not a function or tensor",name);
+			error = 1;
+		}
+		else if ( functions[number].commute == 0 ){
+			MesPrint("&%s is not a noncommuting function or tensor",name);
+			error = 1;
+		}
+		else {
+			*g++ = number+FUNCTION;
+			functions[number].flags |= COULDCOMMUTE;
+			if ( number+FUNCTION >= GAMMA && number+FUNCTION <= GAMMASEVEN ) {
+				functions[GAMMA-FUNCTION].flags |= COULDCOMMUTE;
+				functions[GAMMAI-FUNCTION].flags |= COULDCOMMUTE;
+				functions[GAMMAFIVE-FUNCTION].flags |= COULDCOMMUTE;
+				functions[GAMMASIX-FUNCTION].flags |= COULDCOMMUTE;
+				functions[GAMMASEVEN-FUNCTION].flags |= COULDCOMMUTE;
+			}
+		}
+		*s = c;
+	}
+	return(error);
+}
+
+/*
+  	#] CoCommuteInSet : 
   	#[ CoFunction + ...:
 
 	Function declarations.
@@ -1388,6 +1582,12 @@ retry:;
 			else {
 /*				FUNCTIONS fun = (FUNCTIONS)(AC.Functions->lijst) + numfunction-FUNCTION; */
 				FUNCTIONS fun = (FUNCTIONS)(AC.Functions->lijst) + numfunction;
+
+				if ( fun->tabl != 0 ) {
+					MesPrint("&Illegal attempt to change table into function");
+					error = 1;
+				}
+
 				fun->complex = cplx;
 				fun->commute = comm;
 				if ( istensor && fun->spec == 0 ) {
@@ -1457,11 +1657,14 @@ int DoTable(UBYTE *s, int par)
 {
 	GETIDENTITY
 	UBYTE *name, *p, *inp, c;
-	int i, j, sparseflag = 0, rflag = 0, checkflag = 0, error = 0, ret, oldcbufnum;
-	WORD funnum, type, *OldWork, *w, *newp, *flags1;
+	int i, j, k, sparseflag = 0, rflag = 0, checkflag = 0;
+	int error = 0, ret, oldcbufnum, oldEside;
+	WORD funnum, type, *OldWork, *w, *ww, *t, *tt, *flags1, oldnumrhs,oldnumlhs;
+	LONG oldcpointer;
 	MINMAX *mm, *mm1;
 	LONG x, y;
 	TABLES T;
+	CBUF *C;
 	
 	while ( *s == ',' ) s++;
 	do {
@@ -1484,9 +1687,11 @@ IllForm:	MesPrint("&Illegal name or option in table declaration");
 */
 		if      ( StrICmp(name,(UBYTE *)("check" )) == 0 ) checkflag = 1;
 		else if ( StrICmp(name,(UBYTE *)("zero"  )) == 0 ) checkflag = 2;
+		else if ( StrICmp(name,(UBYTE *)("one"   )) == 0 ) checkflag = 3;
 		else if ( StrICmp(name,(UBYTE *)("strict")) == 0 ) rflag = 1;
 		else if ( StrICmp(name,(UBYTE *)("relax" )) == 0 ) rflag = -1;
-		else if ( StrICmp(name,(UBYTE *)("zerofill" )) == 0 ) rflag = -2;
+		else if ( StrICmp(name,(UBYTE *)("zerofill" )) == 0 ) { rflag = -2; checkflag = 2; }
+		else if ( StrICmp(name,(UBYTE *)("onefill" )) == 0 ) { rflag = -3; checkflag = 3; }
 		else if ( StrICmp(name,(UBYTE *)("sparse")) == 0 ) sparseflag |= 1;
 		else if ( StrICmp(name,(UBYTE *)("base")) == 0 ) sparseflag |= 3;
 		else if ( StrICmp(name,(UBYTE *)("tablebase")) == 0 ) sparseflag |= 3;
@@ -1505,6 +1710,7 @@ IllForm:	MesPrint("&Illegal name or option in table declaration");
 	if ( sparseflag ) {
 		if ( checkflag == 1 ) rflag = 0;
 		else if ( checkflag == 2 ) rflag = -2;
+		else if ( checkflag == 3 ) rflag = -3;
 		else rflag = -1;
 	}
 	if ( ( ret = GetVar(name,&type,&funnum,CFUNCTION,NOAUTO) ) ==
@@ -1537,6 +1743,10 @@ IllForm:	MesPrint("&Illegal name or option in table declaration");
 		MesPrint("&Tensors cannot become tables");
 		return(1);
 	}
+	if ( functions[funnum].symmetric > 0 ) {
+		MesPrint("&Functions with nontrivial symmetrization properties cannot become tables");
+		return(1);
+	}
 	if ( functions[funnum].tabl ) {
 		MesPrint("&Redefinition of an existing table is not allowed.");
 		return(1);
@@ -1593,6 +1803,7 @@ IllForm:	MesPrint("&Illegal name or option in table declaration");
 		T->flags = (WORD *)Malloc1(x*sizeof(WORD),"table flags");
 		mm = T->mm;
 		inp = p;
+		if ( *inp != ')' ) inp++;
 		T->totind = 0;			/* At the moment there are this many */
 		T->tablepointers = 0;
 		T->reserved = 0;
@@ -1637,11 +1848,18 @@ IllForm:	MesPrint("&Illegal name or option in table declaration");
 	Now we redo the 'function part' and send it to the compiler.
 	The prototype has to be picked up properly.
 */
+	AT.WorkPointer++; /* We needs one extra word later */
 	OldWork = AT.WorkPointer;
 	oldcbufnum = AC.cbufnum;
 	AC.cbufnum = T->bufnum;
+	C = cbuf+AC.cbufnum;
+	oldcpointer = C->Pointer - C->Buffer;
+	oldnumlhs = C->numlhs;
+	oldnumrhs = C->numrhs;
+	AddLHS(AC.cbufnum);
 	while ( s >= name ) *--inp = *s--;
-	AC.ProtoType = w = AT.WorkPointer;
+	w = AT.WorkPointer;
+	AC.ProtoType = w;
 	*w++ = SUBEXPRESSION;
 	*w++ = SUBEXPSIZE;
 	*w++ = 0;
@@ -1651,87 +1869,140 @@ IllForm:	MesPrint("&Illegal name or option in table declaration");
 	AC.WildC = w;
 	AC.NwildC = 0;
 	AT.WorkPointer = w + 4*AM.MaxWildcards;
-	AddLHS(AC.cbufnum);
-	if ( ( ret = CompileAlgebra(inp,LHSIDE,AC.ProtoType) ) < 0 ) error = 1;
-	else {
-		if ( AC.NwildC && SortWild(w,AC.NwildC) ) error = 1;
-		w += AC.NwildC;
-		i = w-OldWork;
-		OldWork[1] = i;
-		j = cbuf[AC.cbufnum].Pointer-cbuf[AC.cbufnum].lhs[ret] + T->numind*2-3;
-#ifdef WITHPTHREADS
-		T->prototypeSize = ((i+j)*sizeof(WORD)+2*sizeof(WORD *)) * AM.totalnumberofthreads;
-		T->prototype = (WORD **)Malloc1(T->prototypeSize,"table prototype");
-		T->pattern = T->prototype + AM.totalnumberofthreads;
-		{
-			WORD *t; int k, n;
-			t = (WORD *)(T->pattern + AM.totalnumberofthreads);
-			for ( n = 0; n < AM.totalnumberofthreads; n++ ) {
-				T->prototype[n] = t;
-				for ( k = 0; k < i; k++ ) *t++ = OldWork[k];
-			}
-			T->pattern[0] = t;
-		}
-#else
-		T->prototypeSize = (i+j)*sizeof(WORD);
-		T->prototype = (WORD *)Malloc1(T->prototypeSize, "table prototype");
-		T->pattern = T->prototype + i;
-		while ( --i >= 0 ) T->prototype[i] = OldWork[i];
-#endif
-/*
-		Now check whether wildcards get converted to dollars (for PARALLEL)
-		We give a warning!
-*/
-	{
-		WORD *tw, *twstop;
-#ifdef WITHPTHREADS
-		tw = T->prototype[0];
-#else
-		tw = T->prototype;
-#endif
-		twstop = tw + tw[1]; tw += SUBEXPSIZE;
-		while ( tw < twstop ) {
-			if ( *tw == LOADDOLLAR ) {
-				Warning("The use of $-variable assignments in tables disables parallel\
- execution for the whole program.");
-				AM.hparallelflag |= NOPARALLEL_TBLDOLLAR;
-				AC.mparallelflag |= NOPARALLEL_TBLDOLLAR;
-				AddPotModdollar(tw[2]);
-			}
-			tw += tw[1];
-		}
+	if ( ( ret = CompileAlgebra(inp,LHSIDE,AC.ProtoType) ) < 0 ) {
+		error = 1; goto FinishUp;
 	}
+	if ( AC.NwildC && SortWild(w,AC.NwildC) ) error = 1;
+	w += AC.NwildC;
+	i = w-OldWork;
+	OldWork[1] = i;
+/*
+	Basically we have to pull this pattern through Generator in case
+	there are functions inside functions, or parentheses.
+	We have to temporarily disable the .tabl to avoid problems with
+	TestSub.
+	Essential: we need to start NewSort twice to avoid the PutOut routines.
+	The ground pattern is sitting in C->numrhs, but it could be that it
+	has subexpressions in it. Hence it has to be worked out as the lhs in
+	id statements (in comexpr.c).
+*/
+	OldWork[2] = C->numrhs;
+	*w++ = 1; *w++ = 1; *w++ = 3;
+	OldWork[-1] = w-OldWork+1;
+	AT.WorkPointer = w;
+	ww = C->rhs[C->numrhs];
+	for ( j = 0; j < *ww; j++ ) w[j] = ww[j];
+	AT.WorkPointer = w+*w;
+	if ( *ww == 0 || ww[*ww] != 0 ) {
+		MesPrint("&Illegal table pattern definition");
+		AC.lhdollarflag = 0;
+		error = 1;
+	}
+	if ( error ) goto FinishUp;
 
-		w = cbuf[AC.cbufnum].lhs[ret] + 1;
-#ifdef WITHPTHREADS
-		newp = T->pattern[0];
-#else
-		newp = T->pattern;
-#endif
-		*newp++ = *w++; *newp++ = *w++ + T->numind*2;
-		for ( i = 2; i < FUNHEAD; i++ ) *newp++ = *w++;
-		for ( i = 0; i < T->numind; i++ ) { *newp++ = -SNUMBER; *newp++ = 0; j -= 2; }
-		if ( sparseflag ) { w += 2; j -= 4; }
-		for ( i = FUNHEAD; i < j; i++ ) *newp++ = *w++;
-#ifdef WITHPTHREADS
-		if ( sparseflag ) T->pattern[0][1] = newp - T->pattern[0];
-/*
-		Now we have to copy the pattern for each worker thread
-*/
-		{
-			WORD *t;
-			int k,n;
-			k = newp - T->pattern[0];
-			for ( n = 1; n < AM.totalnumberofthreads; n++ ) {
-				T->pattern[n] = newp; t = T->pattern[0];
-				for ( i = 0; i < k; i++ ) *newp++ = *t++;
-			}
-		}
-#else
-		if ( sparseflag ) T->pattern[1] = newp - T->pattern;
-#endif
+	if ( NewSort(BHEAD0) || NewSort(BHEAD0) ) { error = 1; goto FinishUp; }
+	AN.RepPoint = AT.RepCount + 1;
+	AC.lhdollarflag = 0; oldEside = AR.Eside; AR.Eside = LHSIDE;
+	AR.Cnumlhs = C->numlhs;
+	functions[funnum].tabl = 0;
+	if ( Generator(BHEAD w,C->numlhs) ) {
+		functions[funnum].tabl = T;
+		AR.Eside = oldEside;
+		LowerSortLevel(); LowerSortLevel(); goto FinishUp;
 	}
-	AT.WorkPointer = OldWork;
+	functions[funnum].tabl = T;
+	AR.Eside = oldEside;
+	AT.WorkPointer = w;
+	if ( EndSort(BHEAD w,0) < 0 ) { LowerSortLevel(); goto FinishUp; }
+	if ( *w == 0 || *(w+*w) != 0 ) {
+		MesPrint("&Irregular pattern in table definition");
+		error = 1;
+		goto FinishUp;
+	}
+	LowerSortLevel();
+	if ( AC.lhdollarflag ) {
+		MesPrint("&Unexpanded dollar variables are not allowed in table definition");
+		error = 1;
+		goto FinishUp;
+	}
+	AT.WorkPointer = ww = w + *w;
+	if ( ww[-1] != 3 || ww[-2] != 1 || ww[-3] != 1 ) {
+		MesPrint("&Coefficient of pattern in table definition should be 1.");
+		error = 1;
+		goto FinishUp;
+	}
+	AC.DumNum = 0;
+/*
+	Now we have to allocate space for prototype+pattern
+	In the case of TFORM we need extra pointers, because each worker has its own
+*/
+	j = *w + T->numind*2-3;
+#ifdef WITHPTHREADS
+	{ int n;
+	T->prototypeSize = ((i+j)*sizeof(WORD)+2*sizeof(WORD *)) * AM.totalnumberofthreads;
+	T->prototype = (WORD **)Malloc1(T->prototypeSize,"table prototype");
+	T->pattern = T->prototype + AM.totalnumberofthreads;
+	t = (WORD *)(T->pattern + AM.totalnumberofthreads);
+	for ( n = 0; n < AM.totalnumberofthreads; n++ ) {
+		T->prototype[n] = t;
+		for ( k = 0; k < i; k++ ) *t++ = OldWork[k];
+	}
+	T->pattern[0] = t;
+	j--; w++;
+	w[1] += T->numind*2;
+	for ( k = 0; k < FUNHEAD; k++ ) *t++ = *w++;
+	j -= FUNHEAD;
+	for ( k = 0; k < T->numind; k++ ) { *t++ = -SNUMBER; *t++ = 0; j -= 2; }
+	for ( k = 0; k < j; k++ ) *t++ = *w++;
+	if ( sparseflag ) T->pattern[0][1] = t - T->pattern[0];
+	k = t - T->pattern[0];
+	for ( n = 1; n < AM.totalnumberofthreads; n++ ) {
+		T->pattern[n] = t; tt = T->pattern[0];
+		for ( i = 0; i < k; i++ ) *t++ = *tt++;
+	}
+	}
+#else
+	T->prototypeSize = (i+j)*sizeof(WORD);
+	T->prototype = (WORD *)Malloc1(T->prototypeSize, "table prototype");
+	T->pattern = T->prototype + i;
+	for ( k = 0; k < i; k++ ) T->prototype[k] = OldWork[k];
+	t = T->pattern;
+	j--; w++;
+	w[1] += T->numind*2;
+	for ( k = 0; k < FUNHEAD; k++ ) *t++ = *w++;
+	j -= FUNHEAD;
+	for ( k = 0; k < T->numind; k++ ) { *t++ = -SNUMBER; *t++ = 0; j -= 2; }
+	for ( k = 0; k < j; k++ ) *t++ = *w++;
+	if ( sparseflag ) T->pattern[1] = t - T->pattern;
+#endif
+/*
+	At this point we can pop the compilerbuffer.
+*/
+	C->Pointer = C->Buffer + oldcpointer;
+	C->numrhs = oldnumrhs;
+	C->numlhs = oldnumlhs;
+/*
+	Now check whether wildcards get converted to dollars (for PARALLEL)
+	We give a warning!
+*/
+#ifdef WITHPTHREADS
+	t = T->prototype[0];
+#else
+	t = T->prototype;
+#endif
+	tt = t + t[1]; t += SUBEXPSIZE;
+	while ( t < tt ) {
+		if ( *t == LOADDOLLAR ) {
+			Warning("The use of $-variable assignments in tables disables parallel\
+ execution for the whole program.");
+			AM.hparallelflag |= NOPARALLEL_TBLDOLLAR;
+			AC.mparallelflag |= NOPARALLEL_TBLDOLLAR;
+			AddPotModdollar(t[2]);
+		}
+		t += t[1];
+	}
+FinishUp:;
+	AT.WorkPointer = OldWork - 1;
 	AC.cbufnum = oldcbufnum;
 	if ( T->sparse ) ClearTableTree(T);
 	if ( ( sparseflag & 2 ) != 0 ) {
@@ -1772,6 +2043,50 @@ int CoCTable(UBYTE *s)
 
 /*
   	#] CoCTable : 
+  	#[ EmptyTable :
+*/
+
+void EmptyTable(TABLES T)
+{
+	int j;
+	if ( T->sparse ) ClearTableTree(T);
+	if ( T->boomlijst ) M_free(T->boomlijst,"TableTree");
+	T->boomlijst = 0;
+	for (j = 0; j < T->buffersfill; j++ ) { /* was <= */
+		finishcbuf(T->buffers[j]);
+	}
+	if ( T->buffers ) M_free(T->buffers,"Table buffers");
+	finishcbuf(T->bufnum);
+	T->bufnum = inicbufs();
+	T->bufferssize = 8;
+	T->buffers = (WORD *)Malloc1(sizeof(WORD)*T->bufferssize,"Table buffers");
+	T->buffersfill = 0;
+	T->buffers[T->buffersfill++] = T->bufnum;
+	T->defined = T->mdefined = 0; T->flags = 0;
+	T->numtree = 0; T->rootnum = 0; T->MaxTreeSize = 0;
+	T->spare = 0; T->reserved = 0;
+	if ( T->spare ) {
+		TABLES TT = T->spare;
+		if ( TT->mm ) M_free(TT->mm,"tableminmax");
+		if ( TT->flags ) M_free(TT->flags,"tableflags");
+		if ( TT->tablepointers ) M_free(TT->tablepointers,"tablepointers");
+		for (j = 0; j < TT->buffersfill; j++ ) {
+			finishcbuf(TT->buffers[j]);
+		}
+		if ( TT->boomlijst ) M_free(TT->boomlijst,"TableTree");
+		if ( TT->buffers ) M_free(TT->buffers,"Table buffers");
+		M_free(TT,"table");
+		SpareTable(T);
+	}
+	else {
+		WORD *w = T->tablepointers;
+		j = T->totind;
+		for ( j = TABLEEXTENSION*T->totind; j > 0; j-- ) *w++ = -1; /* means: undefined */
+	}
+}
+
+/*
+  	#] EmptyTable : 
   	#[ AddSet :
 */
 
@@ -1796,6 +2111,7 @@ int AddSet(UBYTE *name, WORD dim)
 	set->last  = AC.SetElementList.num;	/* set has no elements yet */
 	set->type  = -1;					/* undefined as of yet */
 	set->dimension = dim;
+	set->flags = 0;
 	return(numset);
 }
 
@@ -1902,9 +2218,12 @@ int DoElements(UBYTE *s, SETS set, UBYTE *name)
 					MesPrint("&Illegal use of - sign in set. Can use only with vector or number");
 					error = 1;
 				}
+/*
 				numset = AM.OffsetVector - numset;
 				numset |= SPECMASK;
 				numset = AM.OffsetVector - numset;
+*/
+				numset -= WILDMASK;
 			}
 			*s = c;
 			if ( name == 0 && *s == '?' ) {
@@ -1982,6 +2301,15 @@ int DoElements(UBYTE *s, SETS set, UBYTE *name)
 			return(1);
 		}
 	}
+	if ( error == 0 && ( ( set->flags & ORDEREDSET ) == ORDEREDSET ) ) {
+/*
+		The set->last-set->first list of numbers must be sorted.
+		Because we plan here potentially thousands of elements we use
+		a simple version of splitmerge. In ordered sets we can search
+		later with a binary search.
+*/
+		SimpleSplitMerge(SetElements+set->first,set->last-set->first);
+	}
 	return(error);
 }
 
@@ -1994,7 +2322,7 @@ int DoElements(UBYTE *s, SETS set, UBYTE *name)
 
 int CoSet(UBYTE *s)
 {
-	int type, error = 0;
+	int type, error = 0, ordered = 0;
 	UBYTE *name, c, *ss;
 	SETS set;
 	WORD numberofset, dim = MAXPOSITIVE;
@@ -2018,7 +2346,27 @@ IllForm:MesPrint("&Illegal name for set");
 		set = Sets + numberofset;
 		return(0);		/* empty set */
 	}
-	*s = c; ss = s;
+	*s = c; ss = s;     /* ss marks the end of the name */
+	if ( *s == '(' ) {
+		UBYTE *sss, cc;
+		s++; sss = s;   /* Beginning of option */
+		while ( *s != ',' && *s != ')' && *s ) s++;
+		cc = *s; *s = 0;
+		if ( StrICont(sss,(UBYTE *)"ordered") == 0 ) {
+			ordered = ORDEREDSET;
+		}
+		else {
+			MesPrint("&Error: Illegal option in set definition: %s",sss);
+			error = 1;
+		}
+		*s = cc;
+		if ( *s != ')' ) {
+			MesPrint("&Error: Currently only one option allowed in set definition.");
+			error = 1;
+			while ( *s && *s != ')' ) s++;
+		}
+		s++;
+	}
 	if ( *s == '{' ) {
 		s++;
 		if ( ( *s == 'd' || *s == 'D' ) && s[1] == '=' ) {
@@ -2044,6 +2392,7 @@ IllDim:		MesPrint("&Error: Illegal dimension field for set %s",name);
 	numberofset = AddSet(name,dim);
 	*ss = c;
 	set = Sets + numberofset;
+	set->flags |= ordered;
 	if ( *s != ':' ) {
 		MesPrint("&Proper syntax is `Set name:elements'");
 		return(1);
@@ -2337,6 +2686,7 @@ int AddExpression(UBYTE *name, int x, int y)
 	expr->status = x;
 	expr->printflag = y;
 	PUTZERO(expr->onfile);
+	PUTZERO(expr->size);
 	expr->renum = 0;
 	expr->renumlists = 0;
 	expr->hidelevel = 0;
@@ -2854,4 +3204,3 @@ int TestName(UBYTE *name)
 /*
   	#] TestName : 
 */
-
