@@ -5,7 +5,7 @@
  */
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2013 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2017 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -254,8 +254,8 @@ void poly::coefficients_modulo (UWORD *a, WORD na, bool small) {
 
 // converts an integer to a string
 const string int_to_string (WORD x) {
-	char res[20];
-	sprintf (res,"%i",x);
+	char res[41];
+	snprintf (res, 41, "%i",x);
 	return res;
 }
 
@@ -374,20 +374,27 @@ const poly & poly::normalize() {
 
 	// find and sort all monomials
 	// terms[0]/num_vars+3 is an upper bound for number of terms in a
-	WORD **p = (WORD **)Malloc1(terms[0]/(AN.poly_num_vars+3) * sizeof(WORD*), "poly::normalize");
+
+    LONG poffset = AT.pWorkPointer;
+    WantAddPointers((terms[0]/(AN.poly_num_vars+3)));
+    AT.pWorkPointer += terms[0]/(AN.poly_num_vars+3);
+    #define p (&(AT.pWorkSpace[poffset]))
+
+//	WORD **p = (WORD **)Malloc1(terms[0]/(AN.poly_num_vars+3) * sizeof(WORD*), "poly::normalize");
 	
 	int nterms = 0;
 	for (int i=1; i<terms[0]; i+=terms[i])
 		p[nterms++] = terms + i;
 
-	sort(&p[0], &p[nterms], monomial_larger(BHEAD0));
+	sort(p, p + nterms, monomial_larger(BHEAD0));
 
 	WORD *tmp;
-	if (size_of_terms == AM.MaxTer/(LONG)sizeof(WORD))
+	if (size_of_terms == AM.MaxTer/(LONG)sizeof(WORD)) {
 		tmp = (WORD *)TermMalloc("poly::normalize");
-	else
+	}
+	else {
 		tmp = (WORD *)Malloc1(size_of_terms * sizeof(WORD), "poly::normalize");
-		
+	}
 	int j=1;
 	int prevj=0;
 	tmp[0]=0;
@@ -432,13 +439,16 @@ const poly & poly::normalize() {
 	tmp[0] = j;
 	WCOPY(terms,tmp,tmp[0]);
 
-	M_free(p, "poly::normalize");
+//	M_free(p, "poly::normalize");
 	
 	if (size_of_terms == AM.MaxTer/(LONG)sizeof(WORD))
 		TermFree(tmp, "poly::normalize");
 	else
 		M_free(tmp, "poly::normalize");
 
+	AT.pWorkPointer = poffset;
+	#undef p
+
 	return *this;
 }
 
@@ -453,10 +463,6 @@ WORD poly::last_monomial_index () const {
 	return terms[0] - ABS(terms[terms[0]-1]) - AN.poly_num_vars - 2;
 }
 
-/*
-  	#] last_monomial_index : 
-  	#[ last_monomial :
-*/
 
 // pointer to the last monomial (the constant term)
 WORD * poly::last_monomial () const {
@@ -464,7 +470,63 @@ WORD * poly::last_monomial () const {
 }
 
 /*
-  	#] last_monomial : 
+  	#] last_monomial_index : 
+  	#[ compare_degree_vector :
+*/
+
+int poly::compare_degree_vector(const poly & b) const {
+	POLY_GETIDENTITY(*this);
+
+	// special cases if one or both are 0
+	if (terms[0] == 1 &&  b[0] == 1) return 0;
+	if (terms[0] == 1) return -1;
+	if (b[0] == 1) return 1;
+
+	for (int i = 0; i < AN.poly_num_vars; i++) {
+		int d1 = degree(i);
+		int d2 = b.degree(i);
+		if (d1 != d2) return d1 - d2;
+	}
+
+	return 0;
+}
+
+/*
+  	#] compare_degree_vector : 
+  	#[ degree_vector :
+*/
+
+std::vector<int> poly::degree_vector() const {
+	POLY_GETIDENTITY(*this);
+
+	std::vector<int> degrees(AN.poly_num_vars);
+	for (int i = 0; i < AN.poly_num_vars; i++) {
+		degrees[i] = degree(i);
+	}
+
+	return degrees;
+}
+
+/*
+  	#] degree_vector : 
+  	#[ compare_degree_vector :
+*/
+
+int poly::compare_degree_vector(const vector<int> & b) const {
+	POLY_GETIDENTITY(*this);
+
+	if (terms[0] == 1) return -1;
+
+	for (int i = 0; i < AN.poly_num_vars; i++) {
+		int d1 = degree(i);
+		if (d1 != b[i]) return d1 - b[i];
+	}
+
+	return 0;
+}
+
+/*
+  	#] compare_degree_vector : 
   	#[ add :
 */
 
@@ -1289,7 +1351,7 @@ void poly::divmod_one_term (const poly &a, const poly &b, poly &q, poly &r, bool
 	q[0]=qi;
 	r[0]=ri;
 	
-	if (q.modp!=0) NumberFree(ltbinv,"poly::div_one_term");
+	if (q.modp!=0 || ltbinv != NULL) NumberFree(ltbinv,"poly::div_one_term");
 }	
 
 /*
@@ -1471,7 +1533,7 @@ void poly::divmod_univar (const poly &a, const poly &b, poly &q, poly &r, int va
 	NumberFree(s,"poly::div_univar");
 	NumberFree(t,"poly::div_univar");
 
-	if (q.modp!=0) NumberFree(ltbinv,"poly::div_univar");
+	if (q.modp!=0 || ltbinv!=NULL) NumberFree(ltbinv,"poly::div_univar");
 }
 
 /*
@@ -1488,6 +1550,13 @@ void poly::divmod_univar (const poly &a, const poly &b, poly &q, poly &r, int va
  *   the product of the maximum power in all variables is small, a
  *   hash table is used to add equal terms for extra speed.
  *
+ *   If the input flag check_div is set then if the result of any
+ *   coefficient division results in a non-zero remainder (indicating
+ *   that division has failed over the integers) the output flag div_fail
+ *   will be set and the division will be terminated early (q, r
+ *   will be incorrect). If check_div is not set then non-zero remainders
+ *   from coefficient division will be written into r.
+ *
  *   A heap element h is formatted as follows:
  *   - h[0] = index in a
  *   - h[1] = index in b
@@ -1503,10 +1572,11 @@ void poly::divmod_univar (const poly &a, const poly &b, poly &q, poly &r, int va
  *   For details, see M. Monagan, "Polynomial Division using Dynamic
  *   Array, Heaps, and Packed Exponent Vectors"
  */
-void poly::divmod_heap (const poly &a, const poly &b, poly &q, poly &r, bool only_divides) {
+void poly::divmod_heap (const poly &a, const poly &b, poly &q, poly &r, bool only_divides, bool check_div, bool &div_fail) {
 
 	POLY_GETIDENTITY(a);
 
+	div_fail = false;
 	q[0] = r[0] = 1;
 	
 	WORD nmodq=0;
@@ -1514,6 +1584,7 @@ void poly::divmod_heap (const poly &a, const poly &b, poly &q, poly &r, bool onl
 	
 	WORD nltbinv=0;
 	UWORD *ltbinv=NULL;
+	LONG oldpWorkPointer = AT.pWorkPointer;
 	
 	bool both_mod_small=false;
 	
@@ -1527,7 +1598,7 @@ void poly::divmod_heap (const poly &a, const poly &b, poly &q, poly &r, bool onl
 		else {
 			RaisPowCached(BHEAD q.modp,q.modn,&modq,&nmodq);
 		}
-		ltbinv = NumberMalloc("poly::div_heap");
+		ltbinv = NumberMalloc("poly::div_heap-a");
 
 		if (both_mod_small) {
 			WORD ltb = b[b[1]]*b[2+AN.poly_num_vars];
@@ -1542,9 +1613,10 @@ void poly::divmod_heap (const poly &a, const poly &b, poly &q, poly &r, bool onl
 	int nb=b.number_of_terms();
 	WantAddPointers(nb);
 	WORD **heap = AT.pWorkSpace + AT.pWorkPointer;
+	AT.pWorkPointer += nb;
 	
 	for (int i=0; i<nb; i++) 
-		heap[i] = (WORD *) NumberMalloc("poly::div_heap");
+		heap[i] = (WORD *) NumberMalloc("poly::div_heap-b");
 	int nheap = 1;
 	heap[0][0] = 1;
 	heap[0][1] = 0;
@@ -1555,7 +1627,7 @@ void poly::divmod_heap (const poly &a, const poly &b, poly &q, poly &r, bool onl
 	int qi=1, ri=1;
 
 	int s = nb;
-	WORD *t = (WORD *) NumberMalloc("poly::div_heap");
+	WORD *t = (WORD *) NumberMalloc("poly::div_heap-c");
 
 	// insert contains element that still have to be inserted to the heap
 	// (exists to avoid code duplication).
@@ -1695,6 +1767,12 @@ void poly::divmod_heap (const poly &a, const poly &b, poly &q, poly &r, bool onl
 								(UWORD *)&b[2+AN.poly_num_vars], b[b[1]],
 								(UWORD *)&q[qi+1+AN.poly_num_vars], &nq,
 								(UWORD *)&r[ri+1+AN.poly_num_vars], &nr);
+				if(check_div && nr != 0)
+				{
+					// non-zero remainder from coefficient division => result not expressible in integers
+					div_fail = true;
+					break;
+				}
 			}
 			else {
 				// if both polynomials are modulo p^1, use integer calculus
@@ -1748,11 +1826,12 @@ void poly::divmod_heap (const poly &a, const poly &b, poly &q, poly &r, bool onl
 	r[0] = ri;
 
 	for (int i=0; i<nb; i++)
-		NumberFree(heap[i],"poly::div_heap");
+		NumberFree(heap[i],"poly::div_heap-b");
 
-	NumberFree(t,"poly::div_heap");
+	NumberFree(t,"poly::div_heap-c");
 
-	if (q.modp!=0) NumberFree(ltbinv,"poly::div_heap");
+	if (q.modp!=0||ltbinv!=NULL) NumberFree(ltbinv,"poly::div_heap-a");
+	AT.pWorkPointer = oldpWorkPointer;
 }
 
 /*
@@ -1805,7 +1884,8 @@ void poly::divmod (const poly &a, const poly &b, poly &q, poly &r, bool only_div
 		divmod_univar(a,b,q,r,MaX(vara,varb),only_divides);
 	}
 	else {
-		divmod_heap(a,b,q,r,only_divides);
+		bool div_fail;
+		divmod_heap(a,b,q,r,only_divides,false,div_fail);
 	}
 }
 
@@ -1917,7 +1997,7 @@ int poly::first_variable () const {
 	
 	int var = AN.poly_num_vars;
 	for (int j=0; j<var; j++)
-		if (terms[2+j]>0) var=j;
+		if (terms[2+j]>0) return j;
 	return var;
 }
 
@@ -2362,9 +2442,19 @@ void poly::get_variables (PHEAD vector<WORD *> es, bool with_arghead, bool sort_
 	}
 
 	// AN.poly_vars will be deleted in calling functions from polywrap.c
-	if (AN.poly_num_vars > 0) 
-		AN.poly_vars = (WORD *)Malloc1(AN.poly_num_vars*sizeof(WORD), "AN.poly_vars");
-	
+	// For that we use the function poly_free_poly_vars
+	// We add one to the allocation to avoid copying in poly_divmod
+
+	if ( (AN.poly_num_vars+1)*sizeof(WORD) > (size_t)(AM.MaxTer) ) {
+		// This can happen only in expressions with excessively many variables.
+		AN.poly_vars = (WORD *)Malloc1((AN.poly_num_vars+1)*sizeof(WORD), "AN.poly_vars");
+		AN.poly_vars_type = 1;
+	}
+	else {
+		AN.poly_vars = TermMalloc("AN.poly_vars");
+		AN.poly_vars_type = 0;
+	}
+
 	for (int i=0; i<AN.poly_num_vars; i++)
 		AN.poly_vars[i] = vars[i];
 
@@ -2558,11 +2648,95 @@ void poly::poly_to_argument (const poly &a, WORD *res, bool with_arghead) {
 
 /*
   	#] poly_to_argument : 
+  	#[ poly_to_argument_with_den :
+*/
+
+// converts a polynomial class "poly" divided by a number (nden, den) to a form expression
+// cf. poly::poly_to_argument()
+void poly::poly_to_argument_with_den (const poly &a, WORD nden, const UWORD *den, WORD *res, bool with_arghead) {
+
+	POLY_GETIDENTITY(a);
+
+	// special case: a=0
+	if (a[0]==1) {
+		if (with_arghead) {
+			res[0] = -SNUMBER;
+			res[1] = 0;
+		}
+		else {
+			res[0] = 0;
+		}
+		return;
+	}
+
+	if (with_arghead) {
+		res[1] = AN.poly_num_vars>1 ? DIRTYFLAG : 0; // dirty flag
+		for (int i=2; i<ARGHEAD; i++)
+			res[i] = 0;                                // remainder of arghead
+	}
+
+	WORD nden1;
+	UWORD *den1 = (UWORD *)NumberMalloc("poly_to_argument_with_den");
+
+	int L = with_arghead ? ARGHEAD : 0;
+
+	for (int i=1; i!=a[0]; i+=a[i]) {
+
+		res[L]=1; // length
+
+		bool first=true;
+
+		for (int j=0; j<AN.poly_num_vars; j++)
+			if (a[i+1+j] > 0) {
+				if (first) {
+					first=false;
+					res[L+1] = 1; // symbols
+					res[L+2] = 2; // length
+				}
+				res[L+1+res[L+2]++] = AN.poly_vars[j]; // symbol
+				res[L+1+res[L+2]++] = a[i+1+j];  // power
+			}
+
+		if (!first)	res[L] += res[L+2]; // fix length
+
+		// numerator
+		WORD nc = a[i+a[i]-1];
+		WCOPY(&res[L+res[L]], &a[i+a[i]-1-ABS(nc)], ABS(nc));
+
+		// denominator
+		nden1 = nden;
+		WCOPY(den1, den, ABS(nden));
+
+		if (nden != 1 || den[0] != 1) {
+			// remove gcd(num,den)
+			Simplify(BHEAD (UWORD *)&res[L+res[L]], &nc, den1, &nden1);
+		}
+
+		Pack((UWORD *)&res[L+res[L]], &nc, den1, nden1);  // format
+		res[L] += 2*ABS(nc)+1;                            // fix length
+		res[L+res[L]-1] = SGN(nc)*(2*ABS(nc)+1);          // length of coefficient
+		L += res[L];                                      // fix length
+	}
+
+	NumberFree(den1, "poly_to_argument_with_den");
+
+	if (with_arghead) {
+		res[0] = L;
+		// convert to fast notation if possible
+		ToFast(res,res);
+	}
+	else {
+		res[L] = 0;
+	}
+}
+
+/*
+  	#] poly_to_argument_with_den : 
   	#[ size_of_form_notation :
 */
 
 // the size of the polynomial in form notation (without argheads and fast notation)
-int poly::size_of_form_notation () {
+int poly::size_of_form_notation () const {
 	
 	POLY_GETIDENTITY(*this);
 
@@ -2585,6 +2759,36 @@ int poly::size_of_form_notation () {
 
 /*
   	#] size_of_form_notation : 
+  	#[ size_of_form_notation_with_den :
+*/
+
+// the size of the polynomial divided by a number (its size is given by nden)
+// in form notation (without argheads and fast notation)
+// cf. poly::size_of_form_notation()
+int poly::size_of_form_notation_with_den (WORD nden) const {
+
+	POLY_GETIDENTITY(*this);
+
+	// special case: a=0
+	if (terms[0]==1) return 0;
+
+	nden = ABS(nden);
+	int len = 0;
+
+	for (int i=1; i!=terms[0]; i+=terms[i]) {
+		len++;
+		int npow = 0;
+		for (int j=0; j<AN.poly_num_vars; j++)
+			if (terms[i+1+j] > 0) npow++;
+		if (npow > 0) len += 2*npow + 2;
+		len += 2 * MaX(ABS(terms[i+terms[i]-1]), nden) + 1;
+	}
+
+	return len;
+}
+
+/*
+  	#] size_of_form_notation_with_den : 
   	#[ to_coefficient_list :
 */
 
