@@ -4,7 +4,7 @@
  */
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2013 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2017 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -81,7 +81,7 @@ const string factorized_poly::tostring () const {
 	// modulo p^n
 	if (factor[0].modp>0) {
 		res += " (mod ";
-		char tmp[10];
+		char tmp[12];
 		sprintf (tmp,"%i",factor[0].modp);
 		res += tmp;
 		if (factor[0].modn > 1) {
@@ -667,10 +667,13 @@ const vector<poly> polyfact::lift_variables (const poly &A, const vector<poly> &
 	// Calculate the maximum degree of A in x2,...,xm
 	int maxdegA=0;
 	for (int i=1; i<(int)x.size(); i++)
-		maxdegA = MaX(maxdegA, 1+A.degree(x[i]));
+		maxdegA = MaX(maxdegA, A.degree(x[i]));
 
 	// Iteratively add the variables x2,...,xm
 	for (int xi=1; xi<(int)x.size(); xi++) {
+		// replace the current leading coefficients by the correct ones
+		for (int i=0; i<(int)a.size(); i++)
+			a[i] += (lc[i] - a[i].lcoeff_univar(x[0])) * poly::simple_poly(BHEAD x[0],0,a[i].degree(x[0]));
 
 		vector<poly> anew(a);
 		for (int i=0; i<(int)anew.size(); i++)
