@@ -5,7 +5,7 @@
  */
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2013 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2017 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -28,7 +28,7 @@
  *   You should have received a copy of the GNU General Public License along
  *   with FORM.  If not, see <http://www.gnu.org/licenses/>.
  */
-/* #] License : */
+/* #] License : */ 
 /*
   	#[ Includes : sch.c
 */
@@ -59,7 +59,7 @@ static int noextralinefeed = 0;
 static int lowestlevel = 1;
 
 /*
-  	#] Includes :
+  	#] Includes : 
  	#[ schryf-Utilities :
  		#[ StrCopy :			UBYTE *StrCopy(from,to)
 */
@@ -71,7 +71,7 @@ UBYTE *StrCopy(UBYTE *from, UBYTE *to)
 }
 
 /*
- 		#] StrCopy :
+ 		#] StrCopy : 
  		#[ AddToLine :			VOID AddToLine(s)
 
 	Puts the characters of s in the outputline. If the line becomes
@@ -175,7 +175,7 @@ VOID AddToLine(UBYTE *s)
 }
 
 /*
- 		#] AddToLine :
+ 		#] AddToLine : 
  		#[ FiniLine :			VOID FiniLine()
 */
 
@@ -257,7 +257,7 @@ VOID FiniLine()
 }
 
 /*
- 		#] FiniLine :
+ 		#] FiniLine : 
  		#[ IniLine :			VOID IniLine(extrablank)
 
 	Initializes the output line for the type of output
@@ -289,7 +289,7 @@ VOID IniLine(WORD extrablank)
 }
 
 /*
- 		#] IniLine :
+ 		#] IniLine : 
  		#[ LongToLine :			VOID LongToLine(a,na)
 
 	Puts a Long integer in the output line. If it is only a single
@@ -324,7 +324,7 @@ VOID LongToLine(UWORD *a, WORD na)
 }
 
 /*
- 		#] LongToLine :
+ 		#] LongToLine : 
  		#[ RatToLine :			VOID RatToLine(a,na)
 
 	Puts a rational number in the output line. The sign is ignored.
@@ -356,23 +356,35 @@ VOID RatToLine(UWORD *a, WORD na)
 		  }
 		  if ( adenom == 1 && a[na] == 1 ) {
 			LongToLine(a,anumer);
-			if ( anumer > 1 ) { AddToLine((UBYTE *)".D0"); }
+			if ( anumer > 1 ) {
+				if ( AO.DoubleFlag == 2 ) { AddToLine((UBYTE *)".Q0"); }
+				else { AddToLine((UBYTE *)".D0"); }
+			}
 		  }
 		  else if ( anumer == 1 && a[0] == 1 ) {
 			a += na;
 			AddToLine((UBYTE *)"(one/");
 			LongToLine(a,adenom);
-			if ( adenom > 1 ) { AddToLine((UBYTE *)".D0"); }
+			if ( adenom > 1 ) {
+				if ( AO.DoubleFlag == 2 ) { AddToLine((UBYTE *)".Q0"); }
+				else { AddToLine((UBYTE *)".D0"); }
+			}
 			AddToLine((UBYTE *)")");
 		  }
 		  else {
 			if ( anumer > 1 || adenom > 1 ) {
 				LongToLine(a,anumer);
-				if ( anumer > 1 ) { AddToLine((UBYTE *)".D0"); }
+				if ( anumer > 1 ) {
+					if ( AO.DoubleFlag == 2 ) { AddToLine((UBYTE *)".Q0"); }
+					else { AddToLine((UBYTE *)".D0"); }
+				}
 				a += na;
 				AddToLine((UBYTE *)"/");
 				LongToLine(a,adenom);
-				if ( adenom > 1 ) { AddToLine((UBYTE *)".D0"); }
+				if ( adenom > 1 ) {
+					if ( AO.DoubleFlag == 2 ) { AddToLine((UBYTE *)".Q0"); }
+					else { AddToLine((UBYTE *)".D0"); }
+				}
 			}
 			else {
 				AddToLine((UBYTE *)"((one*");
@@ -399,7 +411,8 @@ VOID RatToLine(UWORD *a, WORD na)
 				}
 			}
 			else if ( AC.OutputMode == FORTRANMODE || AC.OutputMode == CMODE ) {
-				if ( AO.DoubleFlag ) { AddToLine((UBYTE *)".D0/"); }
+				if ( AO.DoubleFlag == 2 ) { AddToLine((UBYTE *)".Q0/"); }
+				else if ( AO.DoubleFlag == 1 ) { AddToLine((UBYTE *)".D0/"); }
 				else { AddToLine((UBYTE *)"./"); }
 			}
 			else AddToLine((UBYTE *)"/");
@@ -413,7 +426,8 @@ VOID RatToLine(UWORD *a, WORD na)
 				}
 			}
 			else if ( AC.OutputMode == FORTRANMODE || AC.OutputMode == CMODE ) {
-				if ( AO.DoubleFlag ) { AddToLine((UBYTE *)".D0"); }
+				if ( AO.DoubleFlag == 2 ) { AddToLine((UBYTE *)".Q0"); }
+				else if ( AO.DoubleFlag == 1 ) { AddToLine((UBYTE *)".D0"); }
 				else { AddToLine((UBYTE *)"."); }
 			}
 		  }
@@ -427,7 +441,8 @@ VOID RatToLine(UWORD *a, WORD na)
 					AddToLine((UBYTE *)".");
 				}
 			}
-			else if ( AO.DoubleFlag ) { AddToLine((UBYTE *)".D0"); }
+			else if ( AO.DoubleFlag == 2 ) { AddToLine((UBYTE *)".Q0"); }
+			else if ( AO.DoubleFlag == 1 ) { AddToLine((UBYTE *)".D0"); }
 			else { AddToLine((UBYTE *)"."); }
 		  }
 		  else if ( AC.OutputMode == FORTRANMODE && AC.IsFortran90 == ISFORTRAN90 ) {
@@ -440,7 +455,9 @@ VOID RatToLine(UWORD *a, WORD na)
 		  }
 		  else if ( ( AC.OutputMode == FORTRANMODE || AC.OutputMode == CMODE )
 		  && AO.DoubleFlag ) {
-			AddToLine((UBYTE *)".D0");
+			if ( anumer == 1 && adenom == 1 && a[0] == 1 ) {}
+			else if ( AO.DoubleFlag == 2 ) { AddToLine((UBYTE *)".Q0"); }
+			else if ( AO.DoubleFlag == 1 ) { AddToLine((UBYTE *)".D0"); }
 		  }
 		}
 	}
@@ -504,7 +521,7 @@ VOID RatToLine(UWORD *a, WORD na)
 }
 
 /*
- 		#] RatToLine :
+ 		#] RatToLine : 
  		#[ TalToLine :			VOID TalToLine(x)
 
 	Writes the unsigned number x to the output as a single token.
@@ -529,7 +546,7 @@ VOID TalToLine(UWORD x)
 }
 
 /*
- 		#] TalToLine :
+ 		#] TalToLine : 
  		#[ TokenToLine :		VOID TokenToLine(s)
 
 	Puts s in the output buffer. If it doesn't fit the buffer is
@@ -642,7 +659,7 @@ VOID TokenToLine(UBYTE *s)
 }
 
 /*
- 		#] TokenToLine :
+ 		#] TokenToLine : 
  		#[ CodeToLine :			VOID CodeToLine(name,number,mode)
 
 	Writes a name and possibly its number to output as a single token.
@@ -658,7 +675,32 @@ UBYTE *CodeToLine(WORD number, UBYTE *Out)
 }
 
 /*
- 		#] CodeToLine :
+ 		#] CodeToLine : 
+ 		#[ MultiplyToLine :
+*/
+
+void MultiplyToLine()
+{
+	int i;
+	if ( AO.CurrentDictionary > 0 && AO.CurDictSpecials > 0
+	 && AO.CurDictSpecials == DICT_DOSPECIALS ) {
+		DICTIONARY *dict = AO.Dictionaries[AO.CurrentDictionary-1];
+/*
+		Find the star:
+*/
+		for ( i = 0; i < dict->numelements; i++ ) {
+			if ( dict->elements[i]->type != DICT_SPECIALCHARACTER ) continue;
+			if ( (UBYTE)dict->elements[i]->lhs[0] == (UBYTE)('*') ) {
+				TokenToLine((UBYTE *)(dict->elements[i]->rhs));
+				return;
+			}
+		}
+	}
+	TokenToLine((UBYTE *)"*");
+}
+
+/*
+ 		#] MultiplyToLine : 
  		#[ AddArrayIndex :
 */
 
@@ -678,7 +720,7 @@ UBYTE *AddArrayIndex(WORD num,UBYTE *out)
 }
 
 /*
- 		#] AddArrayIndex :
+ 		#] AddArrayIndex : 
  		#[ PrtTerms :			VOID PrtTerms()
 */
 
@@ -700,7 +742,7 @@ VOID PrtTerms()
 }
 
 /*
- 		#] PrtTerms :
+ 		#] PrtTerms : 
  		#[ WrtPower :
 */
 
@@ -716,7 +758,8 @@ UBYTE *WrtPower(UBYTE *Out, WORD Power)
 		if ( Power < 2*MAXPOWER )
 			Out = NumCopy(Power,Out);
 		else
-			Out = StrCopy(VARNAME(symbols,(LONG)Power-2*MAXPOWER),Out);
+			Out = StrCopy(FindSymbol((WORD)((LONG)Power-2*MAXPOWER)),Out);
+/*			Out = StrCopy(VARNAME(symbols,(LONG)Power-2*MAXPOWER),Out); */
 		if ( AC.OutputMode == CMODE ) *Out++ = ')';
 		*Out = 0;
 	}
@@ -728,7 +771,8 @@ UBYTE *WrtPower(UBYTE *Out, WORD Power)
 		if ( Power > -2*MAXPOWER )
 			Out = NumCopy(-Power,Out);
 		else
-			Out = StrCopy(VARNAME(symbols,(LONG)(-Power)-2*MAXPOWER),Out);
+			Out = StrCopy(FindSymbol((WORD)((LONG)Power-2*MAXPOWER)),Out);
+/*			Out = StrCopy(VARNAME(symbols,(LONG)(-Power)-2*MAXPOWER),Out); */
 		if ( AC.OutputMode >= FORTRANMODE || AC.OutputMode >= PFORTRANMODE ) *Out++ = ')';
 		*Out = 0;
 	}
@@ -736,22 +780,22 @@ UBYTE *WrtPower(UBYTE *Out, WORD Power)
 }
 
 /*
- 		#] WrtPower :
+ 		#] WrtPower : 
  		#[ PrintTime :
 */
 
-void PrintTime()
+void PrintTime(UBYTE *mess)
 {
 	LONG millitime = TimeCPU(1);
 	WORD timepart = (WORD)(millitime%1000);
 	millitime /= 1000;
 	timepart /= 10;
-	MesPrint("Time = %7l.%2i sec",millitime,timepart);
+	MesPrint("At %s: Time = %7l.%2i sec",mess,millitime,timepart);
 }
 
 /*
- 		#] PrintTime :
-  	#] schryf-Utilities :
+ 		#] PrintTime : 
+  	#] schryf-Utilities : 
  	#[ schryf-Writes :
  		#[ WriteLists :			VOID WriteLists()
 
@@ -774,9 +818,11 @@ VOID WriteLists()
 	UBYTE *OutScr, *Out;
 	EXPRESSIONS e;
 	CBUF *C = cbuf+AC.cbufnum;
+	int olddict = AO.CurrentDictionary;
 	skip = &AO.OutSkip;
 	*skip = 0;
 	AO.OutputLine = AO.OutFill = (UBYTE *)AT.WorkPointer;
+	AO.CurrentDictionary = 0;
 	FiniLine();
 	OutScr = (UBYTE *)AT.WorkPointer + ( TOLONG(AT.WorkTop) - TOLONG(AT.WorkPointer) ) /2;
 	if ( AC.CodesFlag || AC.NamesFlag > 1 ) startvalue = 0;
@@ -828,6 +874,7 @@ VOID WriteLists()
 		*skip = 3;
 		FiniLine();
 		for ( i = startvalue; i < j; i++ ) {
+			Out = StrCopy(FindIndex(i+AM.OffsetIndex),OutScr);
 			Out = StrCopy(VARNAME(indices,i),OutScr);
 			if ( indices[i].dimension >= 0 ) {
 				if ( indices[i].dimension != AC.lDefDim ) {
@@ -1004,14 +1051,20 @@ VOID WriteLists()
 							}
 							break;
 						case CVECTOR:
+							Out = OutScr;
+							if ( number < AM.OffsetVector ) {
+								number += WILDMASK;
+								Out = StrCopy((UBYTE *)"-",Out);
+							}
 							if ( number >= AM.OffsetVector + WILDOFFSET ) {
 								Out = StrCopy(VARNAME(vectors,number
-								-AM.OffsetVector-WILDOFFSET),OutScr);
+								-AM.OffsetVector-WILDOFFSET),Out);
 								StrCopy((UBYTE *)"?",Out);
-								TokenToLine(OutScr);
 							}
-							else
-							TokenToLine(VARNAME(vectors,number-AM.OffsetVector));
+							else {
+								Out = StrCopy(VARNAME(vectors,number-AM.OffsetVector),Out);
+							}
+							TokenToLine(OutScr);
 							break;
 						case CFUNCTION:
 							if ( number >= FUNCTION + (WORD)WILDMASK ) {
@@ -1174,6 +1227,16 @@ VOID WriteLists()
 		}
 		FiniLine();
 	}
+	if ( AO.NumDictionaries > 0 ) {
+		for ( i = 0; i < AO.NumDictionaries; i++ ) {
+			WriteDictionary(AO.Dictionaries[i]);
+		}
+		if ( olddict > 0 )
+			MesPrint("\nCurrently dictionary %s is active\n",
+				AO.Dictionaries[olddict-1]->name);
+		else
+			MesPrint("\nCurrently there is no actice dictionary\n");
+	}
 	if ( AC.CodesFlag ) {
 		if ( C->numlhs > 0 ) {
 			TokenToLine((UBYTE *)" Left Hand Sides:");
@@ -1202,10 +1265,127 @@ VOID WriteLists()
 			FiniLine();
 		}
 	}
+	AO.CurrentDictionary = olddict;
 }
 
 /*
- 		#] WriteLists :
+ 		#] WriteLists : 
+ 		#[ WriteDictionary :
+
+	This routine is part of WriteLists and should be called from there.
+*/
+
+void WriteDictionary(DICTIONARY *dict)
+{
+	GETIDENTITY
+	int i, first;
+	WORD *skip, na, *a, spec, *t, *tstop, j;
+	UBYTE str[2], *OutScr, *Out;
+	WORD oldoutputmode = AC.OutputMode, oldoutputspaces = AC.OutputSpaces;
+	WORD oldoutskip = AO.OutSkip;
+	AC.OutputMode = NORMALFORMAT;
+	AC.OutputSpaces = NOSPACEFORMAT;
+	MesPrint("===Contents of dictionary %s===",dict->name);
+	skip = &AO.OutSkip;
+	*skip = 3;
+	AO.OutputLine = AO.OutFill = (UBYTE *)AT.WorkPointer;
+	for ( j = 0; j < *skip; j++ ) *(AO.OutFill)++ = ' ';
+
+	OutScr = (UBYTE *)AT.WorkPointer + ( TOLONG(AT.WorkTop) - TOLONG(AT.WorkPointer) ) /2;
+	for ( i = 0; i < dict->numelements; i++ ) {
+		switch ( dict->elements[i]->type ) {
+			case DICT_INTEGERNUMBER:
+				LongToLine((UWORD *)(dict->elements[i]->lhs),dict->elements[i]->size);
+				Out = OutScr; *Out = 0;
+				break;
+			case DICT_RATIONALNUMBER:
+				a = dict->elements[i]->lhs;
+				na = a[a[0]-1]; na = (ABS(na)-1)/2;
+				RatToLine((UWORD *)(a+1),na);
+				Out = OutScr; *Out = 0;
+				break;
+			case DICT_SYMBOL:
+				na = dict->elements[i]->lhs[0];
+				Out = StrCopy(VARNAME(symbols,na),OutScr);
+				break;
+			case DICT_VECTOR:
+				na = dict->elements[i]->lhs[0]-AM.OffsetVector;
+				Out = StrCopy(VARNAME(vectors,na),OutScr);
+				break;
+			case DICT_INDEX:
+				na = dict->elements[i]->lhs[0]-AM.OffsetIndex;
+				Out = StrCopy(VARNAME(indices,na),OutScr);
+				break;
+			case DICT_FUNCTION:
+				na = dict->elements[i]->lhs[0]-FUNCTION;
+				Out = StrCopy(VARNAME(functions,na),OutScr);
+				break;
+			case DICT_FUNCTION_WITH_ARGUMENTS:
+				t = dict->elements[i]->lhs;
+				na = *t-FUNCTION;
+				Out = StrCopy(VARNAME(functions,na),OutScr);
+				spec = functions[*t - FUNCTION].spec;
+				tstop = t + t[1];
+				first = 1;
+				if ( t[1] <= FUNHEAD ) {}
+				else if ( spec >= TENSORFUNCTION ) {
+					t += FUNHEAD; *Out++ = (UBYTE)'(';
+					while ( t < tstop ) {
+						if ( first == 0 ) *Out++ = (UBYTE)(',');
+						else first = 0;
+						j = *t++;
+						if ( j >= 0 ) {
+							if ( j < AM.OffsetIndex ) { Out = NumCopy(j,Out); }
+							else if ( j < AM.IndDum ) {
+								Out = StrCopy(VARNAME(indices,j-AM.OffsetIndex),Out);
+							}
+							else {
+								MesPrint("Currently wildcards are not allowed in dictionary elements");
+								Terminate(-1);
+							}
+						}
+						else {
+							Out = StrCopy(VARNAME(vectors,j-AM.OffsetVector),Out);
+						}
+					}
+					*Out++ = (UBYTE)')'; *Out = 0;
+				}
+				else {
+					t += FUNHEAD; *Out++ = (UBYTE)'('; *Out = 0;
+					TokenToLine(OutScr);
+					while ( t < tstop ) {
+						if ( !first ) TokenToLine((UBYTE *)",");
+						WriteArgument(t);
+						NEXTARG(t)
+						first = 0;
+					}
+					Out = OutScr;
+					*Out++ = (UBYTE)')'; *Out = 0;
+				}
+				break;
+			case DICT_SPECIALCHARACTER:
+				str[0] = (UBYTE)(dict->elements[i]->lhs[0]);
+				str[1] = 0; 
+				Out = StrCopy(str,OutScr);
+				break;
+			default:
+				Out = OutScr; *Out = 0;
+				break;
+		}
+		Out = StrCopy((UBYTE *)": \"",Out);
+		Out = StrCopy((UBYTE *)(dict->elements[i]->rhs),Out);
+		Out = StrCopy((UBYTE *)"\"",Out);
+		TokenToLine(OutScr);
+		FiniLine();
+	}
+	MesPrint("========End of dictionary %s===",dict->name);
+	AC.OutputMode = oldoutputmode;
+	AC.OutputSpaces = oldoutputspaces;
+	AO.OutSkip = oldoutskip;
+}
+
+/*
+ 		#] WriteDictionary : 
  		#[ WriteArgument :		VOID WriteArgument(WORD *t)
 
 		Write a single argument field. The general field goes to
@@ -1231,6 +1411,8 @@ VOID WriteArgument(WORD *t)
 	}
 	else if ( *t == -SYMBOL ) {
 		if ( t[1] >= MAXVARIABLES-cbuf[AM.sbufnum].numrhs ) {
+			Out = StrCopy(FindExtraSymbol(MAXVARIABLES-t[1]),Out);
+/*
 			Out = StrCopy((UBYTE *)AC.extrasym,Out);
 			if ( AC.extrasymbols == 0 ) {
 				Out = NumCopy((MAXVARIABLES-t[1]),Out);
@@ -1239,6 +1421,7 @@ VOID WriteArgument(WORD *t)
 			else if ( AC.extrasymbols == 1 ) {
 				Out = AddArrayIndex((MAXVARIABLES-t[1]),Out);
 			}
+*/
 /*
 			else if ( AC.extrasymbols == 2 ) {
 				Out = NumCopy((MAXVARIABLES-t[1]),Out);
@@ -1246,17 +1429,20 @@ VOID WriteArgument(WORD *t)
 */
 		}
 		else {
-			StrCopy(VARNAME(symbols,t[1]),Out);
+			StrCopy(FindSymbol(t[1]),Out);
+/*			StrCopy(VARNAME(symbols,t[1]),Out); */
 		}
 	}
 	else if ( *t == -VECTOR ) {
 		if ( t[1] == FUNNYVEC ) { *Out++ = '?'; *Out = 0; }
 		else
-			StrCopy(VARNAME(vectors,t[1] - AM.OffsetVector),Out);
+			StrCopy(FindVector(t[1]),Out);
+/*			StrCopy(VARNAME(vectors,t[1] - AM.OffsetVector),Out); */
 	}
 	else if ( *t == -MINVECTOR ) {
 		*Out++ = '-';
-		StrCopy(VARNAME(vectors,t[1] - AM.OffsetVector),Out);
+		StrCopy(FindVector(t[1]),Out);
+/*		StrCopy(VARNAME(vectors,t[1] - AM.OffsetVector),Out); */
 	}
 	else if ( *t == -INDEX ) {
 		if ( t[1] >= 0 ) {
@@ -1273,14 +1459,16 @@ VOID WriteArgument(WORD *t)
 				}
 				else {
 					i -= AM.OffsetIndex;
-					Out = StrCopy(VARNAME(indices,i%WILDOFFSET),Out);
-						if ( i >= WILDOFFSET ) { *Out++ = '?'; *Out = 0; }
+					Out = StrCopy(FindIndex(i%WILDOFFSET+AM.OffsetIndex),Out);
+/*					Out = StrCopy(VARNAME(indices,i%WILDOFFSET),Out); */
+					if ( i >= WILDOFFSET ) { *Out++ = '?'; *Out = 0; }
 				}
 			}
 		}
 		else if ( t[1] == FUNNYVEC ) { *Out++ = '?'; *Out = 0; }
 		else
-			StrCopy(VARNAME(vectors,t[1] - AM.OffsetVector),Out);
+			StrCopy(FindVector(t[1]),Out);
+/*			StrCopy(VARNAME(vectors,t[1] - AM.OffsetVector),Out); */
 	}
 	else if ( *t == -DOLLAREXPRESSION ) {
 		DOLLARS d = Dollars + t[1];
@@ -1290,8 +1478,12 @@ VOID WriteArgument(WORD *t)
 	else if ( *t == -EXPRESSION ) {
 		StrCopy(EXPRNAME(t[1]),Out);
 	}
+	else if ( *t == -SETSET ) {
+		StrCopy(VARNAME(Sets,t[1]),Out);
+	}
 	else if ( *t <= -FUNCTION ) {
-		StrCopy(VARNAME(functions,-*t-FUNCTION),Out);
+		StrCopy(FindFunction(-*t),Out);
+/*		StrCopy(VARNAME(functions,-*t-FUNCTION),Out); */
 	}
 	else {
 		MesPrint("Illegal function argument while writing");
@@ -1304,7 +1496,7 @@ CleanUp:
 }
 
 /*
- 		#] WriteArgument :
+ 		#] WriteArgument : 
  		#[ WriteSubTerm :		WORD WriteSubTerm(sterm,first)
 
 	Writes a single subterm field to the output line.
@@ -1335,7 +1527,7 @@ WORD WriteSubTerm(WORD *sterm, WORD first)
 						else IniLine(3);
 					if ( first ) TokenToLine((UBYTE *)" ");
 				}
-				if ( !first ) TokenToLine((UBYTE *)"*");
+				if ( !first ) MultiplyToLine();
 				if ( AC.OutputMode == CMODE && t[1] != 1 ) {
 					if ( AC.Cnumpows >= t[1] && t[1] > 0 ) {
 						po = t[1];
@@ -1349,12 +1541,15 @@ WORD WriteSubTerm(WORD *sterm, WORD first)
 					}
 				}
 				if ( *t < NumSymbols ) {
-					Out = StrCopy(VARNAME(symbols,*t),buffer); t++;
+					Out = StrCopy(FindSymbol(*t),buffer); t++;
+/*					Out = StrCopy(VARNAME(symbols,*t),buffer); t++; */
 				}
 				else {
 /*
 					see also routine PrintSubtermList.
 */
+					Out = StrCopy(FindExtraSymbol(MAXVARIABLES-*t),buffer);
+/*
 					Out = StrCopy((UBYTE *)AC.extrasym,buffer);
 					if ( AC.extrasymbols == 0 ) {
 						Out = NumCopy((MAXVARIABLES-*t),Out);
@@ -1363,6 +1558,7 @@ WORD WriteSubTerm(WORD *sterm, WORD first)
 					else if ( AC.extrasymbols == 1 ) {
 						Out = AddArrayIndex((MAXVARIABLES-*t),Out);
 					}
+*/
 /*
 					else if ( AC.extrasymbols == 2 ) {
 						Out = NumCopy((MAXVARIABLES-*t),Out);
@@ -1389,9 +1585,10 @@ WORD WriteSubTerm(WORD *sterm, WORD first)
 						else IniLine(3);
 					if ( first ) TokenToLine((UBYTE *)" ");
 				}
-				if ( !first ) TokenToLine((UBYTE *)"*");
+				if ( !first ) MultiplyToLine();
 
-				Out = StrCopy(VARNAME(vectors,*t - AM.OffsetVector),buffer);
+				Out = StrCopy(FindVector(*t),buffer);
+/*				Out = StrCopy(VARNAME(vectors,*t - AM.OffsetVector),buffer); */
 				t++;
 				if ( AC.OutputMode == MATHEMATICAMODE ) *Out++ = '[';
 				else *Out++ = '(';
@@ -1406,7 +1603,8 @@ WORD WriteSubTerm(WORD *sterm, WORD first)
 						*Out = 0;
 					}
 					else
-						Out = StrCopy(VARNAME(indices,i - AM.OffsetIndex),Out);
+						Out = StrCopy(FindIndex(i),Out);
+/*						Out = StrCopy(VARNAME(indices,i - AM.OffsetIndex),Out); */
 				}
 				else if ( *t == FUNNYVEC ) { *Out++ = '?'; *Out = 0; }
 				else {
@@ -1427,7 +1625,7 @@ WORD WriteSubTerm(WORD *sterm, WORD first)
 						else IniLine(3);
 					if ( first ) TokenToLine((UBYTE *)" ");
 				}
-				if ( !first ) TokenToLine((UBYTE *)"*");
+				if ( !first ) MultiplyToLine();
 			if ( *t >= 0 ) {
 				if ( *t < AM.OffsetIndex ) {
 					TalToLine((UWORD)(*t++));
@@ -1445,14 +1643,16 @@ WORD WriteSubTerm(WORD *sterm, WORD first)
 					}
 					else {
 						i -= AM.OffsetIndex;
-						Out = StrCopy(VARNAME(indices,i%WILDOFFSET),buffer);
+						Out = StrCopy(FindIndex(i%WILDOFFSET+AM.OffsetIndex),buffer);
+/*						Out = StrCopy(VARNAME(indices,i%WILDOFFSET),buffer); */
 						if ( i >= WILDOFFSET ) { *Out++ = '?'; *Out = 0; }
 					}
 					TokenToLine(buffer);
 				}
 			}
 			else {
-				TokenToLine(VARNAME(vectors,*t - AM.OffsetVector)); t++;
+				TokenToLine(FindVector(*t)); t++;
+/*				TokenToLine(VARNAME(vectors,*t - AM.OffsetVector)); t++; */
 			}
 			first = 0;
 			}
@@ -1475,11 +1675,12 @@ WORD WriteSubTerm(WORD *sterm, WORD first)
 						else IniLine(3);
 					if ( first ) TokenToLine((UBYTE *)" ");
 				}
-				if ( !first ) TokenToLine((UBYTE *)"*");
+				if ( !first ) MultiplyToLine();
 				Out = StrCopy((UBYTE *)"d_(",buffer);
 				if ( *t >= AM.OffsetIndex ) {
 					if ( *t < AM.IndDum ) {
-						Out = StrCopy(VARNAME(indices,*t - AM.OffsetIndex),Out);
+						Out = StrCopy(FindIndex(*t),Out);
+/*						Out = StrCopy(VARNAME(indices,*t - AM.OffsetIndex),Out); */
 						t++;
 					}
 					else {
@@ -1497,7 +1698,8 @@ WORD WriteSubTerm(WORD *sterm, WORD first)
 				*Out++ = ',';
 				if ( *t >= AM.OffsetIndex ) {
 					if ( *t < AM.IndDum ) {
-						Out = StrCopy(VARNAME(indices,*t - AM.OffsetIndex),Out);
+						Out = StrCopy(FindIndex(*t),Out);
+/*						Out = StrCopy(VARNAME(indices,*t - AM.OffsetIndex),Out); */
 						t++;
 					}
 					else {
@@ -1524,16 +1726,18 @@ WORD WriteSubTerm(WORD *sterm, WORD first)
 						else IniLine(3);
 					if ( first ) TokenToLine((UBYTE *)" ");
 				}
-				if ( !first ) TokenToLine((UBYTE *)"*");
+				if ( !first ) MultiplyToLine();
 				if ( AC.OutputMode == CMODE && t[2] != 1 )
 					TokenToLine((UBYTE *)"pow(");
-				Out = StrCopy(VARNAME(vectors,*t - AM.OffsetVector),buffer);
+				Out = StrCopy(FindVector(*t),buffer);
+/*				Out = StrCopy(VARNAME(vectors,*t - AM.OffsetVector),buffer); */
 				t++;
 				if ( AC.OutputMode == FORTRANMODE || AC.OutputMode == PFORTRANMODE
 				 || AC.OutputMode == CMODE )
 					*Out++ = AO.FortDotChar;
 				else *Out++ = '.';
-				Out = StrCopy(VARNAME(vectors,*t - AM.OffsetVector),Out);
+				Out = StrCopy(FindVector(*t),Out);
+/*				Out = StrCopy(VARNAME(vectors,*t - AM.OffsetVector),Out); */
 				t++;
 				if ( *t != 1 ) WrtPower(Out,*t);
 				t++;
@@ -1545,7 +1749,7 @@ WORD WriteSubTerm(WORD *sterm, WORD first)
 #if FUNHEAD != 2
 			t += FUNHEAD - 2;
 #endif
-			if ( !first ) TokenToLine((UBYTE *)"*");
+			if ( !first ) MultiplyToLine();
 			if ( AC.OutputMode == CMODE ) TokenToLine((UBYTE *)"pow(");
 			else TokenToLine((UBYTE *)"(");
 			WriteArgument(t);
@@ -1568,7 +1772,7 @@ WORD WriteSubTerm(WORD *sterm, WORD first)
 			TokenToLine((UBYTE *)")");
 			break;
 		case SUBEXPRESSION:
-			if ( !first ) TokenToLine((UBYTE *)"*");
+			if ( !first ) MultiplyToLine();
 			TokenToLine((UBYTE *)"(");
 			t = cbuf[sterm[4]].rhs[sterm[2]];
 			tt = t;
@@ -1599,7 +1803,14 @@ WORD WriteSubTerm(WORD *sterm, WORD first)
 			if ( *sterm < FUNCTION ) {
 			return(MesPrint("Illegal subterm while writing"));
 			}
-			if ( !first ) TokenToLine((UBYTE *)"*");
+			if ( !first ) MultiplyToLine();
+			first = 1;
+			{ UBYTE *tmp;
+				if ( ( tmp = FindFunWithArgs(sterm) ) != 0 ) {
+					TokenToLine(tmp);
+					break;
+				}
+			}
 			t += FUNHEAD-2;
 
 			if ( *sterm == GAMMA && t[-FUNHEAD+1] == FUNHEAD+1 ) {
@@ -1607,7 +1818,8 @@ WORD WriteSubTerm(WORD *sterm, WORD first)
 			}
 			else {
 				if ( *sterm != DUMFUN ) {
-					Out = StrCopy(VARNAME(functions,*sterm - FUNCTION),buffer);
+					Out = StrCopy(FindFunction(*sterm),buffer);
+/*					Out = StrCopy(VARNAME(functions,*sterm - FUNCTION),buffer); */
 				}
 				else { Out = buffer; *Out = 0; }
 				if ( t >= stopper ) {
@@ -1619,9 +1831,11 @@ WORD WriteSubTerm(WORD *sterm, WORD first)
 				*Out = 0;
 				TokenToLine(buffer);
 			}
-			first = 1;
 			i = functions[*sterm - FUNCTION].spec;
 			if ( i >= TENSORFUNCTION ) {
+				int curdict = AO.CurrentDictionary;
+				if ( AO.CurrentDictionary && AO.CurDictNotInFunctions > 0 )
+					AO.CurrentDictionary = 0;
 				t = sterm + FUNHEAD;
 				while ( t < stopper ) {
 					if ( !first ) TokenToLine((UBYTE *)",");
@@ -1631,7 +1845,8 @@ WORD WriteSubTerm(WORD *sterm, WORD first)
 						if ( j < AM.OffsetIndex ) TalToLine((UWORD)(j));
 						else if ( j < AM.IndDum ) {
 							i = j - AM.OffsetIndex;
-							Out = StrCopy(VARNAME(indices,i%WILDOFFSET),buffer);
+							Out = StrCopy(FindIndex(i%WILDOFFSET+AM.OffsetIndex),buffer);
+/*							Out = StrCopy(VARNAME(indices,i%WILDOFFSET),buffer); */
 							if ( i >= WILDOFFSET ) { *Out++ = '?'; *Out = 0; }
 							TokenToLine(buffer);
 						}
@@ -1654,17 +1869,23 @@ WORD WriteSubTerm(WORD *sterm, WORD first)
 						TokenToLine(buffer);					
 					}
 					else {
-						TokenToLine(VARNAME(vectors,j - AM.OffsetVector));
+						TokenToLine(FindVector(j));
+/*						TokenToLine(VARNAME(vectors,j - AM.OffsetVector)); */
 					}
 				}
+				AO.CurrentDictionary = curdict;
 			}
 			else {
+				int curdict = AO.CurrentDictionary;
+				if ( AO.CurrentDictionary && AO.CurDictNotInFunctions > 0 )
+					AO.CurrentDictionary = 0;
 				while ( t < stopper ) {
 					if ( !first ) TokenToLine((UBYTE *)",");
 					WriteArgument(t);
 					NEXTARG(t)
 					first = 0;
 				}
+				AO.CurrentDictionary = curdict;
 			}
 			TokenToLine(closepar);
 			closepar[0] = (UBYTE)')';
@@ -1674,7 +1895,7 @@ WORD WriteSubTerm(WORD *sterm, WORD first)
 }
 
 /*
- 		#] WriteSubTerm :
+ 		#] WriteSubTerm : 
  		#[ WriteInnerTerm :		WORD WriteInnerTerm(term,first)
 
 	Writes the contents of term to the output.
@@ -1728,7 +1949,9 @@ WORD WriteInnerTerm(WORD *term, WORD first)
 				if ( AC.OutputSpaces == NOSPACEFORMAT ) IniLine(1);
 				else IniLine(3);
 		}
-		RatToLine((UWORD *)t,n); first = 0;
+		if ( AO.CurrentDictionary > 0 ) TransformRational((UWORD *)t,n);
+		else                             RatToLine((UWORD *)t,n);
+		first = 0;
 	}
 	else first = 1;
 	while ( s < t ) {
@@ -1793,7 +2016,7 @@ WORD WriteInnerTerm(WORD *term, WORD first)
 				}
 				if ( pow > 1 ) {
 					if ( AC.OutputMode == CMODE ) {
-						if ( !first ) TokenToLine((UBYTE *)"*");
+						if ( !first ) MultiplyToLine();
 						TokenToLine((UBYTE *)"pow(");
 						first = 1;
 					}
@@ -1825,7 +2048,7 @@ WORD WriteInnerTerm(WORD *term, WORD first)
 }
 
 /*
- 		#] WriteInnerTerm :
+ 		#] WriteInnerTerm : 
  		#[ WriteTerm :			WORD WriteTerm(term,lbrac,first,prtf,br)
 
 	Writes a term to output. It tests the bracket information first.
@@ -2151,7 +2374,7 @@ WrtTmes:				t = term;
 }
 
 /*
- 		#] WriteTerm :
+ 		#] WriteTerm : 
  		#[ WriteExpression :	WORD WriteExpression(terms,ltot)
 
 	Writes a subexpression to output.
@@ -2185,7 +2408,7 @@ WORD WriteExpression(WORD *terms, LONG ltot)
 }
 
 /*
- 		#] WriteExpression :
+ 		#] WriteExpression : 
  		#[ WriteAll :			WORD WriteAll()
 
 		Writes all expressions that should be written
@@ -2298,16 +2521,20 @@ WORD WriteAll()
 					if ( !first ) TokenToLine((UBYTE *)",");
 					switch ( n ) {
 						case SYMTOSYM :
-							TokenToLine(VARNAME(symbols,t[2]));
+							TokenToLine(FindSymbol(t[2]));
+/*							TokenToLine(VARNAME(symbols,t[2])); */
 							break;
 						case VECTOVEC :
-							TokenToLine(VARNAME(vectors,t[2] - AM.OffsetVector));
+							TokenToLine(FindVector(t[2]));
+/*							TokenToLine(VARNAME(vectors,t[2] - AM.OffsetVector)); */
 							break;
 						case INDTOIND :
-							TokenToLine(VARNAME(indices,t[2] - AM.OffsetIndex));
+							TokenToLine(FindIndex(t[2]));
+/*							TokenToLine(VARNAME(indices,t[2] - AM.OffsetIndex)); */
 							break;
 						default :
-							TokenToLine(VARNAME(functions,t[2] - FUNCTION));
+							TokenToLine(FindFunction(t[2]));
+/*							TokenToLine(VARNAME(functions,t[2] - FUNCTION)); */
 							break;
 					}
 					t += t[1];
@@ -2401,19 +2628,20 @@ AboWrite:
 }
 
 /*
- 		#] WriteAll :
+ 		#] WriteAll : 
  		#[ WriteOne :			WORD WriteOne(name,alreadyinline)
 
 		Writes one expression from the preprocessor
 */
 
-WORD WriteOne(UBYTE *name, int alreadyinline, int nosemi)
+WORD WriteOne(UBYTE *name, int alreadyinline, int nosemi, WORD plus)
 {
 	GETIDENTITY
 	WORD number;
 	WORD lbrac, first;
 	POSITION pos;
 	FILEHANDLE *f;
+	WORD prf;
 
 	if ( GetName(AC.exprnames,name,&number,NOAUTO) != CEXPRESSION ) {
 		MesPrint("@%s is not an expression",name);
@@ -2457,6 +2685,8 @@ WORD WriteOne(UBYTE *name, int alreadyinline, int nosemi)
 */
 	if ( AR.GetFile == 2 ) f = AR.hidefile;
 	else f = AR.infile;
+	prf = Expressions[number].printflag;
+	if ( plus ) prf |= PRINTONETERM;
 /*
 		Now position the file
 */
@@ -2513,10 +2743,20 @@ WORD WriteOne(UBYTE *name, int alreadyinline, int nosemi)
 				IniLine(0);
 				startinline = alreadyinline;
 				AO.OutFill = AO.OutputLine + startinline;
+				if ( WriteTerm(AO.termbuf,&lbrac,first,0,0) )
+					goto AboWrite;
+				first = 0;
+			}
+			else {
+				if ( ( prf & PRINTONETERM ) != 0 ) first = 1;
+				if ( first ) {
+					FiniLine();
+					IniLine(0);
+				}
+				first = 0;
+				if ( WriteTerm(AO.termbuf,&lbrac,first,0,0) )
+					goto AboWrite;
 			}
-			if ( WriteTerm(AO.termbuf,&lbrac,first,0,0) )
-				goto AboWrite;
-			first = 0;
 		}
 		if ( first ) {
 			IniLine(0);
@@ -2555,9 +2795,6 @@ AboWrite:
 }
 
 /*
- 		#] WriteOne :
-  	#] schryf-Writes :
+ 		#] WriteOne : 
+  	#] schryf-Writes : 
 */
-
-
-
