@@ -7,7 +7,7 @@
  */
 /* #[ License : */
 /*
- *   Copyright (C) 1984-2013 J.A.M. Vermaseren
+ *   Copyright (C) 1984-2017 J.A.M. Vermaseren
  *   When using this file you are requested to refer to the publication
  *   J.A.M.Vermaseren "New features of FORM" math-ph/0010025
  *   This is considered a matter of courtesy as the development was paid
@@ -60,7 +60,7 @@ int tokenize(UBYTE *in, WORD leftright)
 	int error = 0, object, funlevel = 0, bracelevel = 0, explevel = 0, numexp;
 	int polyflag = 0;
 	WORD number, type;
-	UBYTE *s, c;
+	UBYTE *s = in, c;
 	SBYTE *out, *outtop, num[MAXNUMSIZE], *t;
 	LONG i;
 	if ( AC.tokens == 0 ) {
@@ -175,7 +175,7 @@ donumber:		i = 0;
 			case 1:		/* 0-9 */
 				CHECKPOLY
 				s = in;
-				while ( *s == '0' && s[1] == '0' ) s++;
+				while ( *s == '0' && FG.cTable[s[1]] == 1 ) s++;
 				in = s+1; i = 1;
 				while ( FG.cTable[*in] == 1 ) { in++; i++; }
 				if ( object > 0 ) {
@@ -212,7 +212,7 @@ donumber:		i = 0;
 					if ( leftright == LHSIDE ) {
 						if ( object == 1 ) {	/* follows a name */
 							in++; *out++ = TWILDCARD;
-							if ( FG.cTable[in[0]] == 0 || in[0] == '[' ) object = 0;
+							if ( FG.cTable[in[0]] == 0 || in[0] == '[' || in[0] == '{' ) object = 0;
 						}
 						else if ( object == -1 ) {	/* follows comma or ( */
 							in++; s = in;
@@ -230,6 +230,11 @@ donumber:		i = 0;
 							}
 							object = 1;
 						}
+						else {
+							MesPrint("&Illegal position for ?");
+							error = 1;
+							in++;
+						}
 					}
 					else {
 						if ( object != -1 ) goto IllPos;
@@ -248,6 +253,10 @@ donumber:		i = 0;
 							*out++ = (SBYTE)i;
 						}
 						else {
+							if ( AC.vectorlikeLHS == 0 ) {
+								MesPrint("&Generated index ? only allowed in vector substitution",s);
+								error = 1;
+							}
 							*out++ = TGENINDEX;
 						}
 						object = 1;
@@ -467,6 +476,10 @@ donumber:		i = 0;
 			case 7:		/* { | } */
 				CHECKPOLY
 				if ( *in == '{' ) {
+					if ( object > 0 ) {
+						MesPrint("&Illegal position for %s",in);
+						if ( !error ) error = 1;
+					}
 					s = in+1;
 					SKIPBRA2(in)
 					number = DoTempSet(s,in);
@@ -487,7 +500,7 @@ donumber:		i = 0;
 				if ( *in != '!' || leftright == RHSIDE
 				|| object != 1 || out[-1] != TWILDCARD ) goto IllPos;
 				*out++ = TNOT;
-				if ( FG.cTable[in[1]] == 0 || in[1] == '[' ) object = 0;
+				if ( FG.cTable[in[1]] == 0 || in[1] == '[' || in[1] == '{' ) object = 0;
 				in++;
 				break;
 			default:
@@ -730,20 +743,20 @@ int simpwtoken(SBYTE *s)
 			switch ( *t ) {
 				case TSYMBOL:
 				case TDUBIOUS:
-					*w++ = SYMTOSYM; *w++ = 4; *w++ = num; *w++ = -1; break;
+					*w++ = SYMTOSYM; *w++ = 4; *w++ = num; *w++ = num; break;
 				case TINDEX:
 					num += AM.OffsetIndex;
-					*w++ = INDTOIND; *w++ = 4; *w++ = num; *w++ = 0;  break;
+					*w++ = INDTOIND; *w++ = 4; *w++ = num; *w++ = num;  break;
 				case TVECTOR:
 					num += AM.OffsetVector;
-					*w++ = VECTOVEC; *w++ = 4; *w++ = num; *w++ = 0;  break;
+					*w++ = VECTOVEC; *w++ = 4; *w++ = num; *w++ = num;  break;
 				case TFUNCTION:
 					num += FUNCTION;
-					*w++ = FUNTOFUN; *w++ = 4; *w++ = num; *w++ = 0;  break;
+					*w++ = FUNTOFUN; *w++ = 4; *w++ = num; *w++ = num;  break;
 				default:
 					MesPrint("&Illegal type of wildcard in LHS");
 					error = -1;
-					*w++ = SYMTOSYM; *w++ = 4; *w++ = num; *w++ = -1;  break;
+					*w++ = SYMTOSYM; *w++ = 4; *w++ = num; *w++ = num;  break;
 					break;
 			}
 /*
@@ -930,11 +943,13 @@ int simp2token(SBYTE *s)
 						}
 						else if ( ( n == (FIRSTBRACKET-FUNCTION)
 						|| n == (TERMSINEXPR-FUNCTION)
+						|| n == (SIZEOFFUNCTION-FUNCTION)
 						|| n == (NUMFACTORS-FUNCTION)
 						|| n == (GCDFUNCTION-FUNCTION)
 						|| n == (DIVFUNCTION-FUNCTION)
 						|| n == (REMFUNCTION-FUNCTION)
 						|| n == (INVERSEFUNCTION-FUNCTION)
+						|| n == (MULFUNCTION-FUNCTION)
 						|| n == (FACTORIN-FUNCTION)
 						|| n == (FIRSTTERM-FUNCTION)
 						|| n == (CONTENTTERM-FUNCTION) )
@@ -966,14 +981,14 @@ int simp2token(SBYTE *s)
 						}
 						else {
 							if ( *v == TNUMBER || *v == TNUMBER1 ) {
-							  if ( BITSINWORD == 16 ) { LONG x; WORD base;
+							  if ( BITSINWORD == 16 ) { ULONG x; WORD base;
 								base = ( *v == TNUMBER ) ? 100: 128;
 								vv = v+1; x = 0; while ( *vv >= 0 ) { x = x*base + *vv++; }
 								if ( ( vv != t ) || ( ( vv - v ) > 4 ) || ( x > (MAXPOSITIVE+1) ) )
 									*fill++ = *s++;
 								else { *t = TEMPTY; s++; break; }
 							  }
-							  else if ( BITSINWORD == 32 ) { LONG x; WORD base;
+							  else if ( BITSINWORD == 32 ) { ULONG x; WORD base;
 								base = ( *v == TNUMBER ) ? 100: 128;
 								vv = v+1; x = 0; while ( *vv >= 0 ) { x = x*base + *vv++; }
 								if ( ( vv != t ) || ( ( vv - v ) > 6 ) || ( x > (MAXPOSITIVE+1) ) )
@@ -999,6 +1014,7 @@ int simp2token(SBYTE *s)
 							else *fill++ = *s++;
 							break;
 						}
+						/* fall through */
 					case TSETNUM:
 						v++; while ( *v >= 0 ) v++;
 						goto tcommon;
@@ -1007,6 +1023,8 @@ int simp2token(SBYTE *s)
 						|| v[1] == DENOMINATORSYMBOL ) && v[2] < 0 ) {
 							*fill++ = *s++; break;
 						}
+						/* fall through */
+					case TSET:
 					case TVECTOR:
 					case TINDEX:
 					case TFUNCTION:
@@ -1025,14 +1043,14 @@ tcommon:				v++; while ( *v >= 0 ) v++;
 						break;
 					case TNUMBER:
 					case TNUMBER1:
-						if ( BITSINWORD == 16 ) { LONG x; WORD base;
+						if ( BITSINWORD == 16 ) { ULONG x; WORD base;
 							base = ( *v == TNUMBER ) ? 100: 128;
 							vv = v+1; x = 0; while ( *vv >= 0 ) { x = x*base + *vv++; }
 							if ( ( vv != t ) || ( ( vv - v ) > 4 ) || ( x > MAXPOSITIVE ) )
 								*fill++ = *s++;
 							else { *t = TEMPTY; s++; break; }
 						}
-						else if ( BITSINWORD == 32 ) { LONG x; WORD base;
+						else if ( BITSINWORD == 32 ) { ULONG x; WORD base;
 							base = ( *v == TNUMBER ) ? 100: 128;
 							vv = v+1; x = 0; while ( *vv >= 0 ) { x = x*base + *vv++; }
 							if ( ( vv != t ) || ( ( vv - v ) > 6 ) || ( x > MAXPOSITIVE ) )
@@ -1077,7 +1095,8 @@ tcommon:				v++; while ( *v >= 0 ) v++;
 						if ( n == GCDFUNCTION-FUNCTION
 						|| n == DIVFUNCTION-FUNCTION
 						|| n == REMFUNCTION-FUNCTION
-						|| n == INVERSEFUNCTION-FUNCTION ) {
+						|| n == INVERSEFUNCTION-FUNCTION
+						|| n == MULFUNCTION-FUNCTION ) {
 							*t = TEMPTY; s++;
 						}
 						else *fill++ = *s++;
@@ -1712,6 +1731,7 @@ int simp4token(SBYTE *s)
 {
 	int error = 0, n, nsym, settype;
 	WORD i, *w, *wstop, level;
+	SBYTE *const s0 = s;
 	SBYTE *fill = s, *s1, *s2, *s3, type, s1buf[10];
 	SBYTE *tbuf = s, *t, *t1;
 
@@ -1721,8 +1741,8 @@ int simp4token(SBYTE *s)
 			else *fill++ = *s++;
 			continue;
 		}
-		if ( fill[-1] == TWILDCARD ) { *fill++ = *s++; continue; }
-		if ( fill[-1] == TNOT && fill[-2] == TWILDCARD ) { *fill++ = *s++; continue; }
+		if ( fill >= (s0+1) && fill[-1] == TWILDCARD ) { *fill++ = *s++; continue; }
+		if ( fill >= (s0+2) && fill[-1] == TNOT && fill[-2] == TWILDCARD ) { *fill++ = *s++; continue; }
 		s1 = s++; n = 0; while ( *s >= 0 ) { n = 128*n + *s++; }
 		i = Sets[n].type;
 		if ( *s != LBRACE ) { while ( s1 < s ) *fill++ = *s1++; continue; }
@@ -1734,7 +1754,8 @@ int simp4token(SBYTE *s)
 		if ( *s != TSYMBOL && *s != TDOLLAR ) {
 			MesPrint("&Set index in RHS is not a wildcard symbol or $-variable");
 			error = 1;
-			while ( s1 < s ) *fill++ = *s1++; continue;
+			while ( s1 < s ) *fill++ = *s1++;
+			continue;
 		}
 		settype = ( *s == TDOLLAR );
 		s++; nsym = 0; s2 = s;
@@ -1742,7 +1763,8 @@ int simp4token(SBYTE *s)
 		if ( *s != RBRACE ) {
 			MesPrint("&Improper set argument in RHS");
 			error = 1;
-			while ( s1 < s ) *fill++ = *s1++; continue;
+			while ( s1 < s ) *fill++ = *s1++;
+			continue;
 		}
 		s++;
 /*
@@ -1827,7 +1849,8 @@ int simp4token(SBYTE *s)
 					fill--;
 					MesPrint("&Set index in RHS is not a wildcard symbol");
 					error = 1;
-					while ( s1 < s ) *fill++ = *s1++; continue;
+					while ( s1 < s ) *fill++ = *s1++;
+					continue;
 				}
 			}
 		}
@@ -2017,4 +2040,3 @@ int simp6token(SBYTE *tokens, int mode)
  		#] simp6token : 
 	#] Compiler :
 */
-/* temporary commentary for forcing cvs merge */
